# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class Template1LevelMVMixedResponsecc(Template):
    'Fits 1-level multivariate mixed response models (Normal, binary, ordered, or unordered), allowing for missing data, with an option to use multivariate Normal update for beta. Creates custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '1-Level', 'Normal', 'Binary', 'Ordered multinomial', 'Unordered multinomial', 'Multivariate response' ]
    engines = ['CustomC']

    inputs = '''
y = DataMatrix('Responses: ')

for i in range(0, len(y)):
    context['bin'+str(i+1)] = Text('Response Type for response ' + y[i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
    if(context['bin'+str(i+1)] == 'Ordered'):
        context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
    if(context['bin'+str(i+1)] == 'Unordered'):
        context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
        for j in range(0, context['ncats'+y[i]]-1):
            context['x'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
    else:
        context['x'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)

lenbeta = 0
n = 0
for i in range(0, len(y)):
    if(context['bin'+str(i+1)] == 'Normal'):
       context['miss'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Binary'):
       context['misb'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Ordered'):
       context['miso'+y[i]] = ParamVector(monitor=False)
       context['alpha'+y[i]] = ParamVector()
       context['alpha'+y[i]].ncols = -(context['ncats'+y[i]] -1)
       context['alpha'+y[i]].set_value( map(str, range(0, (context['ncats'+y[i]] -1))) )
    if(context['bin'+str(i+1)] == 'Unordered'):
       for j in range(0, context['ncats'+y[i]]-1):
          lenbeta += len(context['x'+str(i+1)+'_'+str(j)])
          context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+ str(j)] = ParamVector(monitor=False)
       context['y'+ str(i+1)] = DataMatrix(value=[])
       context['origv'+str(i+1)] = Text(value = [])
       n += context['ncats'+y[i]] - 1
    else:
       lenbeta += len(context['x'+str(i+1)])
       n += 1
#if n == 1:
#    tau = ParamScalar()
#    sigma = ParamScalar(modelled = False)
#else:
omega_e = ParamMatrix()
omega_e.size = n
omega_e.select('Please input starting values for Level 1 variance matrix: ')
d_e = ParamMatrix(modelled = False)
d_e.size = n
beta = ParamVector()
beta.ncols = lenbeta
deviance = ParamScalar(modelled = False)
'''

    customccode = '''
<%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out

    def PyToC(var):
        if isinstance(var, list):
            return "{" + ",".join(map(PyToC, var)) + "}"
        else:
            return str(var)    
%>
           
<% nvar = len(y) %>\\
<% n = 0 %> \\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n += context['ncats'+y[i]] - 1 %> \\
% else:
<% n += 1 %> \\
% endif
% endfor

% for var in range(0,nvar):
% if (context['bin'+str(var+1)] == 'Ordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% if (context['bin'+str(var+1)] == 'Unordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% endfor
    {
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+y[i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+y[i]]))};

            for(int j = 0; j < length(${y[i]}); j++) {
                if(${y[i]}[j] <= -9.999e29) {
                } else {
                    if(miso${y[i]}[j] < low[int(${y[i]}[j])])
                        low[int(${y[i]}[j])] = miso${y[i]}[j];
                    if(miso${y[i]}[j] > upp[int(${y[i]}[j])])
                        upp[int(${y[i]}[j])] = miso${y[i]}[j];
				}
            } 
            for(int j = 1; j < ncats${y[i]}-1; j++) {
                double ul, ll;            
                if(upp[j]>double(alpha${y[i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${y[i]}[j-1]);
                if(j == (ncats${y[i]}-2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${y[i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${y[i]}[j+1]);
                alpha${y[i]}[j]= ll + (ul-ll) * random01();
            }
        }
% endif
% endfor
        // Bill's attempt of step for d_e
             
        SymMatrix va(${n});
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        va(${i}, ${j}) = omega_e[${count}];
<% count += 1 %>\\
% endfor
% endfor              
        
        bool fail = false;
        mat_d_e = matrix_sym_inverse(va, fail);  
        
        // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb(${n});
        for(int i = 0; i < length(${y[0]}); i++) {
<% lenbeta = 0 %>\\
<% row = 0 %>\\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Normal'):
            double e${row} = double(miss${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)});
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Binary'):
            double e${row} = double(misb${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)});
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
            double e${row} = double(miso${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)});
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
            double e${row} = double(misu${context['ncats'+y[i]]}_${y[i]}_${j}[i]) - (${mmult2(context['x' + str(i+1)+'_'+str(j)], 'beta', 'i', lenbeta)}); 
<% row += 1 %>\\
<% lenbeta += len(context['x' + str(i + 1) + '_'+ str(j)]) %>\\
% endfor
% else :
<% lenbeta += len(context['x' + str(i + 1)]) %>\\
% endif
% endfor
% for i in range(0, n):
% for j in range(i, n):
            sb(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }

        // Now loop over elements of va (omega_e)
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, n):
<% var2 = var1 %>\\
<% withinvar2 = withinvar1 %>\\
% for j in range(i, n):
        {
% if var1 == var2 and (context['bin'+str(var1+1)] != 'Normal'):
            omega_e_accept[${count}] = 44;
% else:
            double llcurr = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
            for(int i1=0; i1<${n}; i1++)
                for(int j1=0; j1<${n}; j1++)
                    llcurr -= sb(i1, j1) * mat_d_e(i1, j1)/2;
            double old = va(${i}, ${j});
            va(${i}, ${j}) = dnorm(va(${i}, ${j}), 1.0/omega_e_sd[${count}]);
            mat_d_e = matrix_sym_inverse(va, fail);
            if (fail == true) {
                va(${i}, ${j}) = old;
                mat_d_e = matrix_sym_inverse(va, fail);
            } else {
                double llnew = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
                for(int i1=0; i1<${n}; i1++)
                    for(int j1=0; j1<${n}; j1++)
                        llnew -= sb(i1, j1) * mat_d_e(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                if(random01() > exp(llnew - llcurr))
                    change = 0;
                }
                if(change == 0) {
                    va(${i}, ${j}) = old;
                    mat_d_e = matrix_sym_inverse(va, fail);
                } else {
                    omega_e_accept[${count}] = int(omega_e_accept[${count}]) + 1;
                }
            }
% endif
<% count += 1 %>\\
        }
% if (context['bin'+str(var2+1)] != 'Unordered'):
<% var2+= 1 %>\\
% elif withinvar2 != (context['ncats'+y[var2]]-2):
<% withinvar2 += 1 %>\\
% else:
<% var2+= 1 %>\\
<% withinvar2 = 0 %>\\
% endif
% endfor
% if (context['bin'+str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+y[var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif
% endfor

<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        omega_e[${count}] = va(${i}, ${j});
        d_e[${count}] = mat_d_e(${i}, ${j});
<% count += 1 %>\\

% endfor
% endfor

                // Step for beta
                std::vector<int> betaind(${lenbeta});
<% betacount = 0 %>
<% row = 0 %>\\
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
% for k in range(0, len(context['x'+str(i+1)+'_'+str(j)])):
                betaind[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endfor
% else:
% for j in range(0, len(context['x' +str(i+1)])):
                betaind[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endif
% endfor

                std::vector<int> betamap(${n});
                int pos = 0;
<% ycount = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
                betamap[${ycount}] = pos;
                pos += mat_x${i + 1}_${j}.cols();
<% ycount += 1 %>
% endfor
% else:
                betamap[${ycount}] = pos;
                pos += mat_x${i + 1}.cols();
<% ycount += 1 %>
% endif
% endfor


                std::vector<double*> tmp_x;
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
% for k in range(0, len(context['x'+str(i+1)+'_'+str(j)])):
                tmp_x.push_back((double*)${context['x' + str(i+1)+'_'+str(j)][k]});
% endfor
% endfor
% else:
% for j in range(0, len(context['x' + str(i+1)])):
                tmp_x.push_back((double*)${context['x' + str(i+1)][j]});
% endfor
% endif
% endfor
                RectMatrix mat_x(tmp_x, length(${y[0]}));

                std::vector<double*> tmp_ymiss;
<%
pyyind = []
pyymap = []
pypos = 0
%>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Normal'):
                tmp_ymiss.push_back(miss${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Binary'):
                tmp_ymiss.push_back(misb${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
                tmp_ymiss.push_back(miso${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
<%
pyymap.append(pypos)
pypos += 1
%>
% for j in range(0,context['ncats'+y[i]]-1):
                tmp_ymiss.push_back(misu${context['ncats'+y[i]]}_${y[i]}_${j});
<% pyyind.append(i)%>
% endfor
% endif
% endfor
                RectMatrix mat_ymiss(tmp_ymiss, length(${y[0]}));

                SymMatrix db(${lenbeta});
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db(j1, j2) += mat_x(i, j1) * mat_d_e(betaind[j1], betaind[j2]) * mat_x(i, j2);
                        }
                    }
                }

                RectMatrix xtdy(${lenbeta}, 1);
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < ${n}; j2++) {
                            xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], j2) * mat_ymiss(i, j2);
                        }
                    }
                }

                fail = false;
                //int err = matrix_sym_invinplace(db);
                //RectMatrix betahat = db * xtdy;
                //mat_beta = dmultnormal(betahat, db, fail);

                RectMatrix betahat = matrix_solve(db, xtdy);
                mat_beta = dmultnormalinv(betahat, db, fail);


% for i in range(0, n):

<%
mat_x_i = 'mat_x' + str(pyyind[i] + 1)
if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
    mat_x_i += '_' + str(i - pyymap[pyyind[i]])
%>

                for (int j = 0; j < mat_ymiss.rows(); j++) {
                    double tmp = 0.0;
% if (context['bin'+str(pyyind[i]+1)] == 'Normal'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
% else:
                    {
% endif
                        for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                            tmp += mat_beta(betamap[${i}] + k, 0) * ${mat_x_i}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e(${i}, k) * mat_ymiss(j, k) * (1.0 / mat_d_e(${i}, ${i}));
                            }
                        }
% for m in range(0, n):
% if i != m:
<%
mat_x_m = 'mat_x' + str(pyyind[m] + 1)
if (context['bin'+str(pyyind[m]+1)] == 'Unordered'):
    mat_x_m += '_' + str(m - pyymap[pyyind[m]])
%>
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1 / mat_d_e(${i}, ${i})) * mat_beta(betamap[${m}] + k, 0) * ${mat_x_m}(j, k);
                        }
% endif
% endfor


% if (context['bin'+str(pyyind[i]+1)] == 'Normal'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    }
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Binary'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) <= 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 1, 0, 0);
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Ordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) == 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if(mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1})
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, alpha${y[i]}[ncats${y[pyyind[i]]}-2], 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 3, alpha${y[i]}[int(mat_y(j, ${i}))-1], alpha${y[i]}[int(mat_y(j, ${i}))]);

% endif

% if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if (mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1}) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if (mat_y(j, ${pyyind[i]}) == ${i - pyymap[pyyind[i]]}) {
                        double limit = 0;
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if ((k != ${i - pyymap[pyyind[i]]}) && (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit))
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, limit, 0);
                    } else {  
                        double limit = mat_ymiss(j, ${pyymap[pyyind[i]]});
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit)
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, limit);
                    }

% endif
                    }
                }
% endfor

    }
'''

    preparedata = '''
mydata = data['datafile']
for i in range(0, len(y)):
    if (context['bin'+str(i+1)] == 'Normal'):
        context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Binary'):
        context['misb'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Ordered'):
        context['miso'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+y[i]]-1):
            context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+str(j)].ncols = -1*len(mydata.variables[y[i]]['data'])
        context['origv'+str(i+1)].name.append(y[i]) 
        for var in context['origv'+str(i+1)].name:
            uniqvals = list(set(mydata.variables[var]['data'].compressed()))
            uniqvals.sort()
            uniqvals.remove(uniqvals[len(uniqvals)-1])
            for j in range(0, len(uniqvals)):
                mydata.addvariable(var + '_' + str(int(j)), data = (mydata.variables[var]['data'][:] == uniqvals[j]).astype(float))
                context['y'+str(i+1)].name.append(var + '_' + str(int(j)))
'''
