# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template


class Template2LevelMissingOnePass(Template):
    'Fits 1 or 2-level multivariate normal and Binary responses with missing data, and a choice of uniform or Wishart priors. Fits a generalised linear model of interest.'

    __version__ = '1.0.1'

    tags = ['Model', '1-Level', 'Normal', 'Binomial', 'Multivariate response']
    engines = ['CustomC']

    inputs = '''
# Here we are combining the templates 1LevelBlockcc for model of interest and 1LevelMVMixedResponsecc for imputation model
# We thus need some renaming to get things correct

moi_y = DataVector('Model of interest (MOI) response: ', help= "<p style='text-align:left'>A.k.a. <em>Y</em>, Outcome variable, Dependent variable, etc.</p>")
moi_D = Text('Specify distribution: ', ['Normal', 'Binomial', 'Poisson', '-ve Binomial'])
if moi_D == 'Binomial':
    moi_n = DataVector('Denominator: ',  help='e.g. if modelling a binary 0/1 response, select a constant of ones.')
    moi_link = Text('Specify link function: ', ['logit', 'probit', 'cloglog'])
if moi_D == 'Poisson':
    moi_link = Text(value = 'ln')
    moi_offset = Boolean('Is there an offset: ', help="An offset allows you to model rates, instead of raw counts.")
    if moi_offset:
        moi_n = DataVector('Offset: ')
if moi_D == '-ve Binomial':
    moi_offset = Boolean('Is there an offset: ')
    if moi_offset:
        moi_n = DataVector('Offset: ')
moi_L2 = Boolean('Do you want a higher level in your model of interest?')
imputation_L2 = Boolean('Do you want a higher level in the imputation model?')
if moi_L2 or imputation_L2:
    L2ID = IDVector('Level 2 ID: ', help="<p style='text-align:left'>Select the variable containing the identifying code for the higher-level groups; the model will then allow the intercept to randomly-vary across these groups.</p>")

moi_numpoly = Integer('Number of polynomial terms in model of interest:', help="<p style='text-align:left'>Note that, for a given variable (e.g. Var1), if you wish to fit Var1^2 and Var1^3 (i.e. quadratic and cubic terms) you would need to request <strong>2</strong> polynomial terms here, and then separately specify each below (i.e. to maximise user control, lower-order powers are not automatically fitted on your behalf).</p>")

moi_terms = list(GUIObject.interface.data['datafile'].variables.keys())

for i in range(0, int(moi_numpoly)):
    context['moi_polyvar'+str(i)] = DataVector('Variable '+str(chr(i + 65)))
    context['moi_polyval'+str(i)] = Text('Power for variable '+str(chr(i + 65)))
    moi_terms.append(str(context['moi_polyvar'+str(i)]) + '^' + str(context['moi_polyval'+str(i)]))

moi_numinter = Integer('Number of interaction terms in model of interest', help="<p style='text-align:left'>You need to specify all interaction terms separately. E.g. if you had a three-way interaction (between VarA, VarB, VarC), and wished to respect marginality and fit the lower-order two-way interactions as well (VarA and VarB; VarA and VarC; VarB and VarC) you would need to request 4 interaction terms, and specify each separately in the subsequent input questions.</p>")
for i in range(0, int(moi_numinter)):
    context['moi_xinter'+str(i)] = List('Variables to include in interaction term ' +str(i + 1), moi_terms)
    t = ''
    for j in range(0, len(context['moi_xinter'+str(i)])):
        if j != 0:
            t += '*'
        t += context['moi_xinter'+str(i)][j]
    moi_terms.append(t)

moi_xvars = List('Explanatory variables for MOI: ', options=moi_terms)

moi_xlist = []
for var in list(moi_xvars):
    if var in GUIObject.interface.data['datafile'].variables.keys():
        moi_xlist.append(var)
moi_x_src = DataMatrix(value=moi_xlist)

if moi_L2:
    moi_x2list = []
    moi_x2vars = List('Explanatory variables with random coefficients in MOI (including intercept): ', options=list(moi_xvars))
    for var in list(moi_x2vars):
        if var in GUIObject.interface.data['datafile'].variables.keys():
            moi_x2list.append(var)
    moi_x2_src = DataMatrix(value=moi_x2list)

imputation_L1resp = Boolean('Do you want lower level responses in the missing data model?')
if imputation_L1resp:
    imputation_y = DataMatrix('Lower level responses with missing data: ', options=list(set(moi_xvars) & set(GUIObject.interface.data['datafile'].variables)), help= "<p style='text-align:left'>A.k.a. <em>Y</em>, Outcome variables, Dependent variables, etc.</p><p style='text-align:left'><strong>Note:</strong> this template assumes responses are continuous or binary.</p><p style='text-align:left'>If you have specified power moi_terms, or interactions, above, these won't be available for you to choose (by design).")
if imputation_L2:
    imputation_L2resp = Boolean('Do you want higher level responses in the imputation model?')
    if imputation_L2resp:
       imputation_L2Data = DataFileChooser('Level 2 data set')
       imputation_y2 = DataMatrix('Higher level responses with missing data: ', options=list(set(moi_xvars) & set(GUIObject.interface.data[str(imputation_L2Data)].variables)), data_src=imputation_L2Data)

if imputation_L1resp:
    for i in range(0, len(imputation_y)):
        context['imputation_resptype'+str(i+1)] = Text('Response Type for response ' + imputation_y[i] + ': ', ['Normal', 'Binary'])
        # Note: removed  allow_cat = True from below; can reinstate if enable categorical explanatory variables to be fitted on user's behalf in MOI
        context['imputation_x'+str(i+1)] = DataMatrix('Explanatory variables for lower level response ' + imputation_y[i] + ' in imputation: ', allow_cat=True, help= "<p style='text-align:left'>A.k.a. X, Predictor variables, Independent variables, etc.</p><p style='text-align:left'><strong>Note:</strong> if you wish to include an <strong>intercept</strong> then you need to add it (e.g. a constant of ones) as one of the explanatory variables.</p><p style='text-align:left'>Once you've selected a variable, you have the opportunity to indicate whether it's categorical or not; if categorical, dummy variables will be added to the model on your behalf.</p>")

if imputation_L2:
    if imputation_L2resp:
        for i in range(0, len(imputation_y2)):
            context['imputation_resptype2_'+str(i+1)] = Text('Response Type for response ' + imputation_y2[i] + ': ', ['Normal', 'Binary'])
            context['imputation_x2_'+str(i+1)] = DataMatrix('Explanatory variables for higher level response ' + imputation_y2[i] + ' in imputation: ', allow_cat=True, data_src=imputation_L2Data, help= "<p style='text-align:left'>A.k.a. X, Predictor variables, Independent variables, etc.</p><p style='text-align:left'><strong>Note:</strong> if you wish to include an <strong>intercept</strong> then you need to add it (e.g. a constant of ones) as one of the explanatory variables.</p><p style='text-align:left'>Once you've selected a variable, you have the opportunity to indicate whether it's categorical or not; if categorical, dummy variables will be added to the model on your behalf.</p>")

omega_e_start = numpy.identity(len(imputation_y))

if imputation_L1resp:
    allnorm = True
    for i in range(0, len(imputation_y)):
        if context['imputation_resptype'+str(i+1)] == 'Binary':
           allnorm = False

    if allnorm:
        imputation_priors = Text('Priors for Level 1 covariance matrix in the imputation model: ', ['Uniform', 'Wishart'])
        if imputation_priors == 'Wishart':
            imputation_R = List('R matrix: ')
            imputation_v = Integer('Degrees of Freedom:')
    else:
        startval = numpy.zeros(int((len(imputation_y)*(len(imputation_y)+1))/2))
        for i in range(0, len(imputation_y)):
            if context['imputation_resptype'+str(i+1)] == 'Normal':
                omega_e_start[i, i] = GUIObject.interface.data['datafile'].variables[imputation_y[i]]['data'].var()
            if context['imputation_resptype'+str(i+1)] == 'Binary':
                omega_e_start[i, i] = 1

if moi_L2:
    moi_priors = Text('Priors for Level 2 random parameters in MOI: ', ['Uniform', 'Wishart'])
    if moi_priors == 'Wishart':
        moi_R = List('R matrix: ')
        moi_v = Integer('Degrees of Freedom:')

if imputation_L2:
    imputation_priors2 = Text('Priors for Level 2 covariance matrix in the imputation model: ', ['Uniform', 'Wishart'])
    if imputation_priors2 == 'Wishart':
        imputation_R2 = List('R matrix: ')
        imputation_v2 = Integer('Degrees of Freedom:')

if moi_L2 or imputation_L2:
    storeresid = Boolean('Store level 2 residuals?')

moi_beta = ParamVector()
moi_beta.ncols = len(moi_xvars)

if moi_D == 'Normal':
    moi_tau = ParamScalar(monitor=False)
    moi_sigma2 = ParamScalar(modelled = False)
if moi_D == '-ve Binomial':
    moi_alpha = ParamScalar()
    moi_rho = ParamVector(parents = [moi_y])

deviance = ParamScalar(modelled = False, monitor=False)

if imputation_L1resp:
    lenbeta = 0
    for i in range(0, len(imputation_y)):
        lenbeta += len(context['imputation_x'+str(i+1)])
        if context['imputation_resptype'+str(i+1)] == 'Normal':
           context['miss'+imputation_y[i]] = ParamVector(monitor=False)
        if context['imputation_resptype'+str(i+1)] == 'Binary':
           context['impb'+imputation_y[i]] = ParamVector(monitor=False)
           context['misb'+imputation_y[i]] = ParamVector(monitor=False)

    imputation_omega_e = ParamMatrix(monitor=True)
    imputation_omega_e.size = len(imputation_y)
    imputation_omega_e.name = [str(i) for i in omega_e_start[numpy.tril_indices(len(imputation_y))]]
    imputation_d_e = ParamMatrix(monitor=False)
    imputation_d_e.size = len(imputation_y)
    imputation_beta = ParamVector(monitor=True)
    imputation_beta.ncols = lenbeta

if moi_L2:
    for i in range(0, len(moi_x2vars)):
        if storeresid:
            context['moi_u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False)
        else:
            context['moi_u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
    moi_omega_u = ParamMatrix(modelled = False)
    moi_omega_u.size = len(moi_x2vars)
    moi_d_u = ParamMatrix(monitor=False)
    moi_d_u.size = len(moi_x2vars)

if imputation_L2:
    lenbeta2 = 0
    num2 = 0
    if imputation_L1resp:
        num2 += len(imputation_y)
    allnorm = True
    if imputation_L2resp:
        num2 += len(imputation_y2)
        for i in range(0, len(imputation_y2)):
            lenbeta2 += len(context['imputation_x2_'+str(i+1)])
            context['imputation_e'+str(i)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
            if context['imputation_resptype2_'+str(i+1)] == 'Normal':
                context['miss'+str(imputation_L2Data)+'_'+imputation_y2[i]] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
                context['miss'+imputation_y2[i]] = ParamVector(monitor=False)
            if context['imputation_resptype2_'+str(i+1)] == 'Binary':
                allnorm = False
                context['impb'+str(imputation_L2Data)+'_'+imputation_y2[i]] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
                context['impb'+imputation_y2[i]] = ParamVector(monitor=False)
                context['misb'+str(imputation_L2Data)+'_'+imputation_y2[i]] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)

        imputation_beta2 = ParamVector(monitor=True)
        imputation_beta2.ncols = lenbeta2

    if imputation_L1resp:
        for i in range(0, len(imputation_y)):
            if storeresid:
                context['imputation_u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False)
            else:
                context['imputation_u' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)

    imputation_omega_u = ParamMatrix(modelled = False)
    imputation_omega_u.size = num2
    imputation_d_u = ParamMatrix(monitor=False)
    imputation_d_u.size = num2


'''

    customccode = '''
<% nresp1 = 0 %>
% if imputation_L1resp:
<% nresp1 += len(imputation_y) %>
% endif
<% nresp2 = 0 %>
% if imputation_L2 and imputation_L2resp:
<% nresp2 += len(imputation_y2) %>
% endif

            std::vector<bool> const_x;
            std::vector<double*> tmp_moi_x;
% for i in range(0, len(moi_xvars)):
% if imputation_L1resp and moi_xvars[i] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(moi_xvars[i])+1)] == 'Normal':
            tmp_moi_x.push_back(miss${moi_xvars[i]});
            const_x.push_back(false);
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(moi_xvars[i])+1)] == 'Binary':
            tmp_moi_x.push_back(impb${moi_xvars[i]});
            const_x.push_back(false);
% endif
% else:
% if imputation_L2 and imputation_L2resp and moi_xvars[i] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(moi_xvars[i])+1)] == 'Normal':
            tmp_moi_x.push_back(miss${moi_xvars[i]});
            const_x.push_back(false);
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(moi_xvars[i])+1)] == 'Binary':
            tmp_moi_x.push_back(impb${moi_xvars[i]});
            const_x.push_back(false);
% endif
% else:
% if moi_xvars[i] in moi_x_src:
            tmp_moi_x.push_back(const_cast<double *>(${moi_x_src[list(moi_x_src).index(moi_xvars[i])]}));
            const_x.push_back(true);
% else:
            std::vector<double> moi_x${i}(length(${moi_x_src[0]}));
            tmp_moi_x.push_back(&moi_x${i}[0]);
            const_x.push_back(false);
% endif
% endif
% endif
% endfor
            RectMatrix mat_moi_x(tmp_moi_x, length(${moi_x_src[0]}));

% if moi_L2:
            std::vector<bool> const_x2;
            std::vector<double*> tmp_moi_x2;
% for i in range(0, len(moi_x2vars)):
% if imputation_L1resp and moi_x2vars[i] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(moi_x2vars[i])+1)] == 'Normal':
            tmp_moi_x2.push_back(miss${moi_x2vars[i]});
            const_x2.push_back(false);
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(moi_x2vars[i])+1)] == 'Binary':
            tmp_moi_x2.push_back(impb${moi_x2vars[i]});
            const_x2.push_back(false);
% endif
% elif imputation_L2 and imputation_L2resp and moi_x2vars[i] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(moi_x2vars[i])+1)] == 'Normal':
            tmp_moi_x2.push_back(miss${moi_x2vars[i]});
            const_x2.push_back(false);
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(moi_x2vars[i])+1)] == 'Binary':
            tmp_moi_x2.push_back(impb${moi_x2vars[i]});
            const_x2.push_back(false);
% endif
% else:
% if moi_x2vars[i] in moi_x2_src:
            tmp_moi_x2.push_back(const_cast<double *>(${moi_x2_src[list(moi_x2_src).index(moi_x2vars[i])]}));
            const_x2.push_back(true);
% else:
            std::vector<double> moi_x2${i}(length(${moi_x2vars_src[0]}));
            tmp_moi_x2.push_back(const_cast<double *>(&moi_x2${i}[0]));
            const_x2.push_back(false);
% endif
% endif
% endfor
            RectMatrix mat_moi_x2(tmp_moi_x2, length(${moi_x2_src[0]}));
% endif

% if imputation_L1resp:
            std::vector<double*> tmp_imputation_x_lev1;
% for i in range(0, nresp1):
% for j in range(0, len(context['imputation_x' + str(i+1)])):
            tmp_imputation_x_lev1.push_back(const_cast<double *>(${context['imputation_x' + str(i+1)][j]}));
% endfor
% endfor
            RectMatrix mat_imputation_x_lev1(tmp_imputation_x_lev1, length(${imputation_y[0]}));
% endif

% if imputation_L2 and imputation_L2resp:
            std::vector<double*> tmp_imputation_x_lev2;
% for i in range(0, nresp2):
% for j in range(0, len(context['imputation_x2_' + str(i+1)])):
            tmp_imputation_x_lev2.push_back(const_cast<double *>(${imputation_L2Data}_${context['imputation_x2_' + str(i+1)][j]}));
% endfor
% endfor
            RectMatrix mat_imputation_x_lev2(tmp_imputation_x_lev2, length(${imputation_L2Data}_${imputation_y2[0]}));
% endif

% if imputation_L1resp:
            std::vector<double*> tmp_ymiss;
% for i in range(0, nresp1):
% if context['imputation_resptype'+str(i+1)] == 'Normal':
            tmp_ymiss.push_back(miss${imputation_y[i]});
% endif
% if context['imputation_resptype'+str(i+1)] == 'Binary':
            tmp_ymiss.push_back(misb${imputation_y[i]});
% endif
% endfor
            RectMatrix mat_imputation_ymiss(tmp_ymiss, length(${imputation_y[0]}));
% endif

% if imputation_L2:
% if imputation_L2resp:
            std::vector<double*> tmp_ymiss2;
% for i in range(0, nresp2):
% if context['imputation_resptype2_'+str(i+1)] == 'Normal':
            tmp_ymiss2.push_back(miss${imputation_L2Data}_${imputation_y2[i]});
% endif
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
            tmp_ymiss2.push_back(misb${imputation_L2Data}_${imputation_y2[i]});
% endif
% endfor
            RectMatrix mat_imputation_ymiss2(tmp_ymiss2, length(${imputation_L2Data}_${imputation_y2[0]}));
% endif



% if imputation_L1resp:
            std::vector<double*> tmp_imputation_u;
% for i in range(0, nresp1):
            tmp_imputation_u.push_back(imputation_u${i});
% endfor
            RectMatrix mat_imputation_u(tmp_imputation_u, length(imputation_u0));
% endif
% endif


% if moi_L2:
            std::vector<double*> tmp_moi_u;
% for i in range(0, len(moi_x2vars)):
            tmp_moi_u.push_back(moi_u${i});
% endfor
            RectMatrix mat_moi_u(tmp_moi_u, length(moi_u0));

            std::vector<long int*> mat_moi_u_accept;
% for i in range(0, len(moi_x2vars)):
            mat_moi_u_accept.push_back(moi_u${i}_accept);
% endfor

            std::vector<double*> tmp_moi_u_sd;
% for i in range(0, len(moi_x2vars)):
            tmp_moi_u_sd.push_back(moi_u${i}_sd);
% endfor
            RectMatrix mat_moi_u_sd(tmp_moi_u_sd, length(moi_u0));
% endif

% if imputation_L2:
            std::vector<double*> imputation_tmp_l2e;
% if imputation_L2resp:
% for i in range(0, nresp2):
            imputation_tmp_l2e.push_back(imputation_e${i});
% endfor
% endif
% if imputation_L1resp:
            int imputation_l2rows = length(imputation_u0);
% for i in range(0, nresp1):
            imputation_tmp_l2e.push_back(imputation_u${i});
% endfor
% else:
            int imputation_l2rows = mat_imputation_ymiss2.rows();
% endif
            RectMatrix imputation_mat_l2e(imputation_tmp_l2e, imputation_l2rows);
% endif

            // Record starting position of beta for each response
% if imputation_L1resp:
            std::vector<int> betamap(${nresp1});
            int pos = 0;
% for i in range(0, nresp1):
            betamap[${i}] = pos;
            pos += mat_imputation_x${i + 1}.cols();
% endfor

            std::vector<int> betamap2(${nresp2});
            int pos2 = 0;
<% ycount = 0 %>
% for i in range(0, nresp2):
            betamap2[${ycount}] = pos2;
            pos2 += mat_imputation_x2_${i+1}.cols();
<% ycount += 1 %>
% endfor

            // Record which response each beta belongs to
            std::vector<int> betaind(mat_imputation_x_lev1.cols());
<% betacount = 0 %>
% for i in range(0, nresp1):
% for j in range(0, len(context['imputation_x' +str(i+1)])):
            betaind[${betacount}] = ${i};
<% betacount += 1 %>
% endfor
% endfor

% if imputation_L2 and imputation_L2resp:
            std::vector<int> betaind2(mat_imputation_x_lev2.cols());
<% betacount2 = 0 %>
% for i in range(0, nresp2):
% for j in range(0, len(context['imputation_x2_' + str(i+1)])):
            betaind2[${betacount2}] = ${i};
<% betacount2 += 1 %> \\
% endfor
% endfor
% endif

            auto recalc_x = [&](auto j) {
% for i1 in range(0, len(moi_xvars)):
% if not moi_xvars[i1] in imputation_y and not moi_xvars[i1] in moi_x_src:
                mat_moi_x(j, ${i1}) = \\
<% xterms = moi_xvars[i1].split('*') %>\\
% for j1 in range(0, len(xterms)):
% if j1 != 0:
 *\\
% endif
<% xpoly = xterms[j1].split('^') %>\\
% if len(xpoly) > 1:
 pow(\\
% endif
% if imputation_L1resp and xpoly[0] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(xpoly[0])+1)] == 'Normal':
 miss${xpoly[0]}[j]\\
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(xpoly[0])+1)] == 'Binary':
 impb${xpoly[0]}[j]\\
% endif
% elif imputation_L2 and imputation_L2resp and xpoly[0] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(xpoly[0])+1)] == 'Normal':
 miss${xpoly[0]}[j]\\
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(xpoly[0])+1)] == 'Binary':
 impb${xpoly[0]}[j]\\
% endif
% else:
 ${xpoly[0]}[j]\\
% endif
% if len(xpoly) > 1:
, ${xpoly[1]})\\
% endif
% endfor
;
% endif
% endfor
            };

            auto recalc_x2 = [&](auto j) {
% if moi_L2:
% for i1 in range(0, len(moi_x2vars)):
% if not moi_x2vars[i1] in imputation_y and not moi_x2vars[i1] in moi_x2_src:
                mat_moi_x2(j, ${i1}) = \\
<% x2terms = moi_x2vars[i1].split('*') %>\\
% for j1 in range(0, len(x2terms)):
% if j1 != 0:
 *\\
% endif
<% x2poly = x2terms[j1].split('^') %>\\
% if len(x2poly) > 1:
 pow(\\
% endif
% if imputation_L1resp and x2poly[0] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(x2poly[0])+1)] == 'Normal':
 miss${x2poly[0]}[j]\\
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(x2poly[0])+1)] == 'Binary':
 impb${x2poly[0]}[j]\\
% endif
% elif imputation_L2 and imputation_L2resp and x2poly[0] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(x2poly[0])+1)] == 'Normal':
 miss${x2poly[0]}[j]\\
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(x2poly[0])+1)] == 'Binary':
 impb${x2poly[0]}[j]\\
% endif
% else:
 ${x2poly[0]}[j]\\
% endif
% if len(x2poly) > 1:
, ${x2poly[1]})\\
% endif
% endfor
;
% endif
% endfor
% endif
            };

            {

                bool fail = false;

                if (runstate == 0) {
% if imputation_L1resp:
% for i in range(0, len(imputation_y)):
                    RunningStat ystats_${i};
                    for (int j = 0; j < length(${imputation_y[i]}); ++j) {
                        if (${imputation_y[i]}[j] > -9.999e29) {
                            ystats_${i}.Push(${imputation_y[i]}[j]);
                        }
                    }
% if context['imputation_resptype'+str(i+1)] == 'Normal':
                    for (int j = 0; j < length(${imputation_y[i]}); ++j) {
                        if (${imputation_y[i]}[j] <= -9.999e29) {
                            miss${imputation_y[i]}[j] = dnormsd(ystats_${i}.Mean(), ystats_${i}.StandardDeviation());
                        }
                    }
% endif
% if context['imputation_resptype'+str(i+1)] == 'Binary':
                    for (int j = 0; j < length(${imputation_y[i]}); ++j) {
                        if (${imputation_y[i]}[j] <= -9.999e29) {
                            impb${imputation_y[i]}[j] = dbin(ystats_${i}.Mean(), ystats_${i}.NumDataValues());
                        }
                        if (impb${imputation_y[i]}[j] == 0) {
                            misb${imputation_y[i]}[j] = dtnormal(0, 1, 2, 0, 0);
                        } else {
                            misb${imputation_y[i]}[j] = dtnormal(0, 1, 1, 0, 0);
                        }
                    }
% endif
% endfor
% endif

% if imputation_L2 and imputation_L2resp:
% for i in range(0, len(imputation_y2)):
                    RunningStat ystats2_${i};
                    for (int j = 0; j < length(${imputation_L2Data}_${imputation_y2[i]}); ++j) {
                        if (${imputation_L2Data}_${imputation_y2[i]}[j] > -9.999e29) {
                            ystats2_${i}.Push(${imputation_L2Data}_${imputation_y2[i]}[j]);
                        }
                    }
% if context['imputation_resptype2_'+str(i+1)] == 'Normal':
                    for (int j = 0; j < length(${imputation_L2Data}_${imputation_y2[i]}); ++j) {
                        if (${imputation_L2Data}_${imputation_y2[i]}[j] <= -9.999e29) {
                            miss${imputation_L2Data}_${imputation_y2[i]}[j] = dnormsd(ystats2_${i}.Mean(), ystats2_${i}.StandardDeviation());
                        }
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                            int k = idmap_${L2ID}[j][id_k];
                            miss${imputation_y2[i]}[k] = miss${imputation_L2Data}_${imputation_y2[i]}[j];
                        }
                    }
% endif
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
                    for (int j = 0; j < length(${imputation_L2Data}_${imputation_y2[i]}); ++j) {
                        if (${imputation_L2Data}_${imputation_y2[i]}[j] <= -9.999e29) {
                            impb${imputation_L2Data}_${imputation_y2[i]}[j] = dbin(ystats2_${i}.Mean(), ystats2_${i}.NumDataValues());
                        }
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                            int k = idmap_${L2ID}[j][id_k];
                            impb${imputation_y2[i]}[k] = impb${imputation_L2Data}_${imputation_y2[i]}[j];
                        }
                        if (impb${imputation_L2Data}_${imputation_y2[i]}[j] == 0) {
                            misb${imputation_L2Data}_${imputation_y2[i]}[j] = dtnormal(0, 1, 2, 0, 0);
                        } else {
                            misb${imputation_L2Data}_${imputation_y2[i]}[j] = dtnormal(0, 1, 1, 0, 0);
                        }
                    }
% endif
% endfor
% endif

% if imputation_L2:
% if imputation_L1resp:
                    for (int i = 0; i < length(imputation_u0); ++i) {
% for i in range(0, nresp1):
                        imputation_u${i}[i] = dnorm(0, 1);
% endfor
                    }
% endif
% endif
                }

                for (int i = 0; i < length(${moi_x_src[0]}); i++) {
                    recalc_x(i);
                }

% if moi_L2:
                for (int i = 0; i < length(${moi_x2_src[0]}); i++) {
                    recalc_x2(i);
                }
% endif

% if moi_L2:
                RectMatrix moi_sb = mat_moi_u.T() * mat_moi_u;

% if moi_priors == 'Uniform':
                int vwint = mat_moi_u.rows() - ${len(moi_x2vars) + 1};
                if (runstate == 0) {
% for j in range(0, len(moi_x2vars)):
                    moi_sb(${j}, ${j}) += 0.0001;
% endfor
                }
% endif
% if moi_priors == 'Wishart':
                int vwint = mat_moi_u.rows() + ${moi_v};
% endif

% if moi_priors == 'Wishart':
<%
import numpy
n = len(moi_x2_src)
moi_Rmat = numpy.empty([n, n])
count = 0 
for j in range(0, n):
    for k in range(0, j + 1):
        moi_Rmat[j, k] = float(moi_R.name[count]) 
        moi_Rmat[k, j] = moi_Rmat[j, k]
        count += 1
%>

% for j in range(0, len(moi_x2vars)):
% for k in range(0, len(moi_x2vars)):
                moi_sb(${j}, ${k}) += ${str(moi_Rmat[j, k] * float(moi_v))};
% endfor
% endfor
% endif
                matrix_sym_invinplace(moi_sb);
                fail = false;
                mat_moi_d_u = dwishart(vwint, moi_sb, fail); 
                if (fail == true) {
                    std::cout << "Wishart failed" << std::endl;
                }
                fail = false;
                mat_moi_omega_u = matrix_sym_inverse(mat_moi_d_u, fail);
                if (fail == true) {
                    std::cout << "Inverse failed" << std::endl;
                }
                for (int i = 0; i < ${len(moi_x2vars) * (len(moi_x2vars) + 1) / 2}; ++i) {
                    moi_d_u_accept[i] = 44;
                    moi_omega_u_accept[i] = 44;
                }

% endif

                fail = false;
% if moi_D == 'Normal':
                RectMatrix moi_ydiff(mat_moi_y);
% if moi_L2:
                for (int p = 0; p < mat_moi_u.cols(); ++p) {
                    for (int i = 0; i < mat_moi_u.rows(); ++i) {
                        double mean = 0.0;
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[i]; ++id_k) {
                            int k = idmap_${L2ID}[i][id_k];
                            double rowmean = 0;
                            rowmean += mat_moi_y(k, 0);
                            for (int j = 0; j < mat_moi_beta.rows(); j++) {
                                rowmean -= mat_moi_x(k, j) * mat_moi_beta(j, 0);
                            }
                            for (int j = 0; j < mat_moi_u.cols(); j++) {
                                if (p != j) {
                                    rowmean -= mat_moi_x2(k, j) * mat_moi_u(i, j);
                                }
                            }
                            mean += mat_moi_x2(k, p) * rowmean;
                        }

                        double moi_v = 0;
                        for (int j = 0; j < mat_moi_u.cols(); ++j) {
                            if (j != p) {
                                moi_v -= mat_moi_d_u(p, j) * mat_moi_u(i, j);
                            }
                        }

                        double x2sq = 0;
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[i]; ++id_k) {
                            int k = idmap_${L2ID}[i][id_k];
                            x2sq += pow(mat_moi_x2(k, p), 2);
                        }
                    
                        double du = mat_moi_d_u(p, p) + moi_tau * x2sq;
                        mat_moi_u(i, p) = dnorm((moi_v + (moi_tau * mean)) / du, du);
                        mat_moi_u_accept[p][i] = 44;
                    }
                }

                for (int i = 0; i < moi_ydiff.rows(); ++i) {
                    for (int j = 0; j < ${len(moi_x2vars)}; ++j) {
                        moi_ydiff(i, 0) -= mat_moi_x2(i, j) * mat_moi_u(${L2ID}[i], j);
                    }
                }
% endif

                // Setting up constant terms for beta step
                RectMatrix xtxchol = matrix_cholesky(mat_moi_x.T() * mat_moi_x, false, fail);
                RectMatrix mean = matrix_cholsolve(xtxchol, mat_moi_x.T() * moi_ydiff);

                // Multivariate step for beta (Model of interest)
                DiagMatrix taudiag = identity(${len(moi_xvars)}) / moi_tau;
                SymMatrix variance = matrix_cholsolve(xtxchol, taudiag);
                mat_moi_beta = dmultnormal(mean, variance, fail);
                for (int i = 0; i < mat_moi_beta.rows(); ++i) {
                    moi_beta_accept[i] = 44;
                }

                // Note: ydiff is y - u from above
                RectMatrix yxb = moi_ydiff - mat_moi_x * mat_moi_beta;

                moi_tau = dgamma(0.001 + 0.5 * mat_moi_y.rows(), 0.001 + 0.5 * (yxb.T() * yxb)(0,0));
                moi_tau_accept = 44;

                moi_sigma2 = 1/moi_tau;
                moi_sigma2_accept = 44;
                deviance = (moi_tau*(yxb.T() * yxb)(0,0)) + (log(3.14159265)-log(moi_tau))*length(${moi_y}) + (2.0*0.346573590279973*length(${moi_y})); // Last term is log(sqrt(2))
                deviance_accept = 44;
% else:
% if moi_L2:
                auto ll_moi_u = [&](auto r, auto c) {
                    double ll = 0.0;

                    for (auto id_i = 0; id_i < length(idmap_${L2ID}[r]); id_i++) {
                        auto i = idmap_${L2ID}[r][id_i];
                        double xb = 0;
                        for (auto j = 0; j < mat_moi_x.cols(); ++j) {
                            xb += mat_moi_x(i, j) * mat_moi_beta(j, 0);
                        }
                        for (auto j = 0; j < ${len(moi_x2vars)}; ++j) {
                            xb += mat_moi_x2(i, j) * mat_moi_u(r, j);
                        }
% if moi_D == 'Binomial':
% if moi_link == 'logit':
                        double temp = -log1p(exp(-xb));
                        ll += (mat_moi_n(i, 0) - mat_moi_y(i, 0)) * (-xb + temp) + mat_moi_y(i, 0) * temp;
% endif
% if moi_link == 'probit':
                        double temp = normprob(-xb);
                        ll += (mat_moi_n(i, 0) - mat_moi_y(i, 0)) * log(temp) + mat_moi_y(i, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
                        double temp = -exp(xb);
                        ll += (mat_moi_n(i, 0) - mat_moi_y(j, 0)) * temp + mat_moi_y(i, 0) * log(1.0 - exp(temp));
% endif
% endif
% if moi_D == 'Poisson':
% if moi_offset:
                        xb += mat_moi_n(i, 0);
% endif
                        ll += mat_moi_y(i, 0) * xb - exp(xb);
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
                        xb += mat_moi_n(i, 0);
% endif
                        ll += mat_moi_y(i, 0) * xb - exp(xb) * mat_moi_rho(i, 0);
% endif
                    }
                    llcurr -= 0.5 * mat_moi_d_u(c, c) * pow(mat_moi_u(r, c), 2);

                    double moi_v = 0;
                    for (auto i = 0; i < mat_moi_u.cols(); ++i) {
                        if (i != c) {
                            moi_v -= mat_moi_d_u(c, i) * mat_moi_u(r, i);
                        }
                    }
                    ll += moi_v * mat_moi_u(r, c);
                    return ll;
                };

                for (int p = 0; p < mat_moi_u.cols(); ++p) {
                    for (int i = 0; i < mat_moi_u.rows(); ++i) {
                        double llcurr = ll_moi_u(i, p);

                        double old = mat_moi_u(i, p);
                        mat_moi_u(i, p) = dnorm(mat_moi_u(i, p), 1.0/mat_moi_u_sd(i, p));

                        double llnew = mat_moi_u(i, p);

                        bool change = true;
                        if (isFinite(llnew) == 0) {
                            change = false;
                        } else {
                            if (llnew > llcurr) {
                                ;
                            } else {
                                double rand = random01();
                                if (rand > exp(llnew - llcurr))
                                    change = false;
                            }
                        }
                        if (change == false) {
                            mat_moi_u(i, p) = old;
                        } else {
                            mat_moi_u_accept[p][i]++;
                        }
                    }
                }
% endif

                auto ll_moi_beta = [&]() {
                    double ll = 0.0;
                    for (auto i = 0; i < mat_moi_x.rows(); ++i) {
                        double xb = 0;
                        for (auto j = 0; j < mat_moi_x.cols(); ++j) {
                            xb += mat_moi_x(i, j) * mat_moi_beta(j, 0);
                        }
% if moi_L2:
                        for (auto j = 0; j < ${len(moi_x2vars)}; ++j) {
                            xb += mat_moi_x2(i, j) * mat_moi_u(${L2ID}[i], j);
                        }
% endif
% if moi_D == 'Binomial':
% if moi_link == 'logit':
                        double temp = -log1p(exp(-xb));
                        ll += (mat_moi_n(i, 0) - mat_moi_y(i, 0)) * (-xb + temp) + mat_moi_y(i, 0) * temp;

% endif
% if moi_link == 'probit':
                        double temp = normprob(-xb);
                        ll += (mat_moi_n(i, 0) - mat_moi_y(i, 0)) * log(temp) + mat_moi_y(i, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
                        double temp = -exp(xb);
                        ll += (mat_moi_n(i, 0) - mat_moi_y(i, 0)) * temp + mat_moi_y(i, 0) * log(1.0 - exp(temp));
% endif
% endif
% if moi_D == 'Poisson':
% if moi_offset:
                        xb += mat_moi_n(i, 0);
% endif
                        llcurr += mat_moi_y(i, 0) * xb - exp(xb);
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
                        xb += mat_moi_n(i, 0);
% endif
                        ll += mat_moi_y(i, 0) * xb - exp(xb) * mat_moi_rho(i, 0);
% endif
                    }
                    return ll;
                };

                for (int i = 0; i < mat_moi_x.cols(); ++i) {
                    double llcurr = ll_moi_beta();

                    double old = mat_moi_beta(i, 0);

                    mat_moi_beta(i, 0) = dnorm(mat_moi_beta(i, 0), 1.0/moi_beta_sd[i]);

                    double llnew = ll_moi_beta();

                    bool change = true;
                    if (isFinite(llnew) == 0) {
                        change = false;
                    } else {
                        if (llnew > llcurr) {
                            ;
                        } else {
                            double rand = random01();
                            if (rand > exp(llnew - llcurr))
                                change = false;
                        }
                    }
                    if (change == false) {
                        mat_moi_beta(i, 0) = old;
                    } else {
                        moi_beta_accept[i]++;
                    }
                }

% if moi_D == '-ve Binomial':
                {
                    double sum = 0.0;
                    for (int j = 0; j < mat_moi_x.rows(); ++j) {
                        sum += mat_moi_rho(j, 0) - log(mat_moi_rho(j, 0));
                    }

                    double llcurr = (((-sum * moi_alpha) + (mat_moi_x.rows() * moi_alpha * log(moi_alpha)) - (0.0005 * pow(log(moi_alpha), 2)) - log(moi_alpha)) - mat_moi_x.rows() * lgamma(moi_alpha));

                    double old = moi_alpha;
                    moi_alpha = dnorm(moi_alpha, 1.0/moi_alpha_sd);

                    double llnew = (((-sum * moi_alpha) + (mat_moi_x.rows() * moi_alpha * log(moi_alpha)) - (0.0005 * pow(log(moi_alpha), 2)) - log(moi_alpha)) - mat_moi_x.rows() * lgamma(moi_alpha));

                    bool change = true;
                    if (isFinite(llnew) == 0) {
                        change = false;
                    } else {
                        if (llnew > llcurr) {
                            ;
                        } else {
                            double rand = random01();
                            if (rand > exp(llnew - llcurr))
                                change = false;
                        }
                    }
                    if (change == false) {
                        moi_alpha = old;
                    } else {
                        moi_alpha_accept++;
                    }

                    for (int i = 0; i < mat_moi_rho.rows(); ++i) {
                        double xb = 0;
                        for (int k = 0; k < mat_moi_x.cols(); ++k) {
                            xb += mat_moi_x(i, k) * mat_moi_beta(k, 0);
                        }
                        mat_moi_rho(i, 0) = dgamma(moi_alpha + mat_moi_y(i, 0), moi_alpha + exp(xb));
                    }
                }
% endif

                deviance = 0.0;
                for (int j = 0; j < mat_moi_x.rows(); ++j) {
                    double xb = 0;
                    for (int k = 0; k < mat_moi_x.cols(); ++k) {
                        xb += mat_moi_x(j, k) * mat_moi_beta(k, 0);
                    }
% if moi_L2:
                    for (int k = 0; k < ${len(moi_x2vars)}; ++k) {
                        xb += mat_moi_x2(j, k) * mat_moi_u(${L2ID}[j], k);
                    }
% endif
% if moi_D == 'Binomial':
% if moi_link == 'logit':
                    double temp = -log1p(exp(-xb));
                    deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * (-xb + temp) + mat_moi_y(j, 0) * temp;
% endif
% if moi_link == 'probit':
                    double temp = normprob(-xb);
                    deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * log(temp) + mat_moi_y(j, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
                    double temp = -exp(xb);
                    deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * temp + mat_moi_y(j, 0) * log(1.0 - exp(temp));
% endif
                    deviance += logfact(mat_moi_n(j, 0)) - logfact(mat_moi_y(j, 0)) - logfact(mat_moi_n(j, 0) - mat_moi_y(j, 0));
% endif
% if moi_D == 'Poisson':
% if moi_offset:
                    xb += mat_moi_n(j, 0);
% endif
                    deviance += mat_moi_y(j, 0) * xb - exp(xb);
                    deviance += logfact(mat_moi_y(j, 0));
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
                    xb += mat_moi_n(j, 0);
% endif
                    deviance += mat_moi_y(j, 0) * xb - exp(xb) * mat_moi_rho(j, 0);
                    deviance -= logfact(mat_moi_y(j, 0));
% endif
                }
                deviance *= -2;
                deviance_accept = 44;
% endif

<% allnorm = True %>
% if imputation_L1resp:
% for i in range(0, nresp1):
% if context['imputation_resptype'+str(i+1)] == 'Binary':
<% allnorm = False %>
% endif
% endfor

% if allnorm:

                // Note currently using a uniform prior for variance matrix
                SymMatrix imputation_sb(${nresp1});
                for(int i = 0; i < mat_moi_y.rows(); i++) {
% for i in range(0, nresp1):
                    double imputation_e${i} = mat_imputation_ymiss(i, ${i});
                    for (int j = 0; j < ${len(context['imputation_x' +str(i+1)])}; j++) {
                        imputation_e${i} -= imputation_beta[betamap[${i}] + j] * mat_imputation_x_lev1(i, betamap[${i}] + j);
                    }
% if imputation_L2:
                    imputation_e${i} -= mat_imputation_u(${L2ID}[i], ${i});
% endif
% endfor

% for i in range(0, nresp1):
% for j in range(0, i + 1):
                    imputation_sb(${i}, ${j}) += imputation_e${i} * imputation_e${j};
% endfor
% endfor
                }
% if imputation_priors == 'Uniform':
                if (runstate == 0) {
% for i in range(0, nresp1):
                    imputation_sb(${i}, ${i}) += 0.0001;
% endfor
                }
% endif
% if imputation_priors == 'Wishart':
<%
import numpy
imputation_Rmat = numpy.empty([nresp1, nresp1])
count = 0 
for i in range(0, nresp1):
    for j in range(0, i + 1):
        imputation_Rmat[i, j] = float(imputation_R[count]) 
        imputation_Rmat[j, i] = imputation_Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, nresp1):
% for j in range(0, i + 1):
                imputation_sb(${i}, ${j}) += ${str(imputation_Rmat[i, j] * float(imputation_v))};
% endfor
% endfor
% endif
                matrix_sym_invinplace(imputation_sb);

% if imputation_priors == 'Uniform':
                int vw = mat_moi_y.rows() - ${nresp1 + 1};
% endif
% if imputation_priors == 'Wishart':
                int vw = mat_moi_y.rows() + ${imputation_v};
% endif
               
                mat_imputation_d_e = dwishart(vw, imputation_sb, fail);
                mat_imputation_omega_e = matrix_sym_inverse(mat_imputation_d_e, fail);
                for (int i = 0; i < ${len(imputation_y) * (len(imputation_y) + 1) / 2}; ++i) {
                    imputation_d_e_accept[i] = 44;
                    imputation_omega_e_accept[i] = 44;
                }
% else:

                // Bill's attempt of step for imputation_d_e

                fail = false;
                mat_imputation_d_e = matrix_sym_inverse(mat_imputation_omega_e, fail);  
        
                // Note currently using a uniform prior for variance matrix
                // Note sb contains all terms needed in MH steps
                SymMatrix imputation_sb(${nresp1});
                for (int i = 0; i < length(${imputation_y[0]}); i++) {
% for i in range(0, nresp1):
                    double imputation_e${i} = mat_imputation_ymiss(i, ${i});
                    for (int j = 0; j < ${len(context['imputation_x' +str(i+1)])}; j++) {
                        imputation_e${i} -= imputation_beta[betamap[${i}] + j] * mat_imputation_x_lev1(i, betamap[${i}] + j);
                    }
% if imputation_L2:
                    imputation_e${i} -= mat_imputation_u(${L2ID}[i], ${i});
% endif
% endfor
% for i in range(0, nresp1):
% for j in range(i, nresp1):
                    imputation_sb(${i}, ${j}) += imputation_e${i} * imputation_e${j};
% endfor
% endfor
                }

                // Now loop over elements of va (imputation_omega_e)
<% count = 0 %>\\
% for i in range(0, nresp1):
% for j in range(0, i + 1):
                {
% if i == j and (context['imputation_resptype'+str(i+1)] != 'Normal'):
                    imputation_omega_e_accept[${count}] = 44;
% else:
                    auto ll_imputation_omega_e = [&](const SymMatrix &d_e) {
                        double ll = length(${imputation_y[0]}) * log(matrix_determinant(d_e)) / 2;
                        for (auto i = 0; i < ${nresp1}; ++i) {
                            for (auto j = 0; j < ${nresp1}; ++j) {
                                ll -= imputation_sb(i, j) * d_e(i, j) / 2;
                            }
                        }
                        return ll;
                    };

                    double llcurr = ll_imputation_omega_e(mat_imputation_d_e);

                    double old = mat_imputation_omega_e(${i}, ${j});
                    mat_imputation_omega_e(${i}, ${j}) = dnorm(mat_imputation_omega_e(${i}, ${j}), 1.0 / imputation_omega_e_sd[${count}]);
                    SymMatrix temp_mat_imputation_d_e = matrix_cholesky(mat_imputation_omega_e, fail);
                    if (fail == true) {
                        mat_imputation_omega_e(${i}, ${j}) = old;
                    } else {
                        temp_mat_imputation_d_e = matrix_cholinverse(temp_mat_imputation_d_e);
                        double llnew = ll_imputation_omega_e(temp_mat_imputation_d_e);

                        bool change = true;
                        if (isFinite(llnew) == 0) {
                            change = false;
                        } else if (llnew > llcurr) {
                            ;
                        } else {
                            if (random01() > exp(llnew - llcurr)) {
                                change = false;
                            }
                        }
                        if (change == false) {
                            mat_imputation_omega_e(${i}, ${j}) = old;
                        } else {
                            imputation_omega_e_accept[${count}]++;
                            imputation_d_e_accept[${count}]++;
                            mat_imputation_d_e = temp_mat_imputation_d_e;
                        }
                    }
% endif
<% count += 1 %>\\
                }
% endfor
% endfor

% endif


                // Step for imputation_beta (imputation)
                SymMatrix imputation_db(mat_imputation_x_lev1.cols());
                for (int i = 0; i < length(${imputation_y[0]}); i++) {
                    for (int j1 = 0; j1 < mat_imputation_x_lev1.cols(); j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            imputation_db(j1, j2) += mat_imputation_x_lev1(i, j1) * mat_imputation_d_e(betaind[j1], betaind[j2]) * mat_imputation_x_lev1(i, j2);
                        }
                    }
                }

                RectMatrix imputation_xtdy(mat_imputation_x_lev1.cols(), 1);
                for (int i = 0; i < length(${imputation_y[0]}); i++) {
                    for (int j1 = 0; j1 < mat_imputation_x_lev1.cols(); j1++) {
                        for (int j2 = 0; j2 < ${nresp1}; j2++) {
                            imputation_xtdy(j1, 0) += mat_imputation_x_lev1(i, j1) * mat_imputation_d_e(betaind[j1], j2) * (mat_imputation_ymiss(i, j2)\\
% if imputation_L2:
- mat_imputation_u(${L2ID}[i], j2)\\
% endif
);
                        }
                    }
                }

                fail = false;

               
                RectMatrix betahat = matrix_solve(imputation_db, imputation_xtdy);
                mat_imputation_beta = dmultnormalinv(betahat, imputation_db, fail);
                for (int i = 0; i < mat_imputation_beta.rows(); ++i) {
                    imputation_beta_accept[i] = 44;
                }
% endif


% if imputation_L2resp:

                // Step for imputation_beta2
                // Calculate submatrix of imputation_omega_u and invert
                SymMatrix mat_imputation_omega_beta(imputation_omega_u, mat_imputation_ymiss2.cols());
                SymMatrix mat_imputation_d_beta = matrix_sym_inverse(mat_imputation_omega_beta, fail);

                SymMatrix imputation_db2(mat_imputation_x_lev2.cols());
                for (int i = 0; i < imputation_l2rows; i++) {
                    for (int j1 = 0; j1 < mat_imputation_x_lev2.cols(); j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            imputation_db2(j1, j2) += mat_imputation_x_lev2(i, j1) * mat_imputation_d_beta(betaind2[j1], betaind2[j2]) * mat_imputation_x_lev2(i, j2);
                        }
                    }
                }

                RectMatrix imputation_xtdy2(mat_imputation_x_lev2.cols(), 1);
                for (int i = 0; i < imputation_l2rows; i++) {
                    for (int j1 = 0; j1 < mat_imputation_x_lev2.cols(); j1++) {
                        for (int j2 = 0; j2 < mat_imputation_ymiss2.cols(); j2++) {
                            imputation_xtdy2(j1, 0) += mat_imputation_x_lev2(i, j1) * mat_imputation_d_beta(betaind2[j1], j2) * mat_imputation_ymiss2(i, j2);
                        }
                    }
                }

                fail = false;
                RectMatrix imputation_betahat2 = matrix_solve(imputation_db2, imputation_xtdy2);
                mat_imputation_beta2 = dmultnormalinv(imputation_betahat2, imputation_db2, fail);
                for (int i = 0; i < mat_imputation_beta2.rows(); ++i) {
                    imputation_beta2_accept[i] = 44;
                }
% endif

% if imputation_L2:
% if imputation_L2resp:
                int off2 = ${nresp2};
% else:
                int off2 = 0;
% endif
                for (int i = 0; i < mat_imputation_u0.rows(); i++) {
                    SymMatrix imputation_duj(${nresp1});
                    for (int j = 0; j < ${nresp1}; j++) {
                        for (int k = 0; k < ${nresp1}; k++) {
                            imputation_duj(j, k) = mat_imputation_d_u(j+off2, k+off2) + mat_imputation_d_e(j, k) * idmap_${L2ID}[i].size();
                        }
                    }

                    RectMatrix imputation_ymxb(${nresp1}, 1);
                    for (int id_k = 0; id_k < idmap_${L2ID}[i].size(); id_k++) {
                        int k = idmap_${L2ID}[i][id_k];
% for j in range(0, nresp1):
                        imputation_ymxb(${j}, 0) += mat_imputation_ymiss(k, ${j});
                        for (int m = 0; m < mat_imputation_x${j + 1}.cols(); m++) {
                            imputation_ymxb(${j}, 0) -= mat_imputation_x${j + 1}(k, m) * mat_imputation_beta(betamap[${j}] + m, 0);
                        }
% endfor
                    }

                    RectMatrix imputation_xtdu1(${nresp1}, 1);
% if imputation_L2resp:
                    std::vector<double> imputation_sumymxb2(${nresp2});
                    for (int m = 0; m < mat_imputation_ymiss2.cols(); m++) {
                        imputation_sumymxb2[m] = mat_imputation_ymiss2(i, m);
                    }
                    for (int m = 0; m < mat_imputation_x_lev2.cols(); m++) {
                        imputation_sumymxb2[betaind2[m]] -= mat_imputation_beta2(m, 0) * mat_imputation_x_lev2(i, m); 
                    }
                    for (int j1 = 0; j1 < ${nresp1}; j1++) {
                        for (int j2 = 0; j2 < ${nresp2}; j2++) {
                            imputation_xtdu1(j1, 0) += mat_imputation_d_u(${nresp2}+j1, j2) * imputation_sumymxb2[j2];
                        }
                    }
% endif


                    imputation_xtdu1 += mat_imputation_d_e * imputation_ymxb;

                    RectMatrix imputation_newuj = dmultnormalinv(matrix_solve(imputation_duj, imputation_xtdu1), imputation_duj, fail);
                    for (int j = 0; j < ${nresp1}; j++) {
                        mat_imputation_u(i, j) = imputation_newuj(j, 0);
                    }
 % for r in range(0, nresp1):
                    imputation_u${r}_accept[i] = 44;
 % endfor

                }
% endif
% endif

% if imputation_L2:

% if imputation_L2resp:
% for i in range(0, nresp2):
                for (int i = 0; i < imputation_l2rows; i++) {
                    imputation_mat_l2e(i, ${i}) = mat_imputation_ymiss2(i, ${i});
                    for (int j = 0; j < mat_imputation_x2_${i+1}.cols(); j++) {
                        imputation_mat_l2e(i, ${i}) -= mat_imputation_x2_${i+1}(i, j) * mat_imputation_beta2(betamap2[${i}] + j, 0);
                    }
                    imputation_e${i}_accept[i] = 44;
                }
% endfor
% endif

                SymMatrix imputation_sbu = imputation_mat_l2e.T() * imputation_mat_l2e;

<% allnorm2 = True %>
% for i in range(0, nresp2):
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
<% allnorm2 = False %>
% endif
% endfor

% if allnorm2:

% if imputation_priors2 == 'Uniform':
                if (runstate == 0) {
                    for (int i = 0; i < imputation_sbu.rows(); i++) {
                        imputation_sbu(i, i) += 0.0001;
                    }
                }
% endif

% if imputation_priors2 == 'Wishart':
<%
import numpy
imputation_Rmat2 = numpy.empty([nresp1+nresp2, nresp1+nresp2])
count = 0 
for i in range(0, nresp1+nresp2):
    for j in range(0, i + 1):
        imputation_Rmat2[i, j] = float(imputation_R2[count]) 
        imputation_Rmat2[j, i] = imputation_Rmat2[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, nresp1+nresp2):
% for j in range(0, i + 1):
                imputation_sbu(${i}, ${j}) += ${str(imputation_Rmat2[i, j] * float(imputation_v2))};
% endfor
% endfor
% endif

                matrix_sym_invinplace(imputation_sbu);
% if imputation_priors2 == 'Uniform':
                int imputation_vwu = imputation_l2rows - ${nresp1+nresp2 + 1};
% endif
% if imputation_priors2 == 'Wishart':
                int imputation_vwu = imputation_l2rows + ${imputation_v2};
% endif
                mat_imputation_d_u = dwishart(imputation_vwu, imputation_sbu, fail);
                mat_imputation_omega_u = matrix_sym_inverse(mat_imputation_d_u, fail);
                for (int i = 0; i < mat_imputation_d_u.rows() * (mat_imputation_d_u.rows() + 1) / 2; ++i) {
                    imputation_d_u_accept[i] = 44;
                }
                for (int i = 0; i < mat_imputation_omega_u.rows() * (mat_imputation_omega_u.rows() + 1) / 2; ++i) {
                    imputation_omega_u_accept[i] = 44;
                }
% else:
                // Need to add binary MH step
                if (runstate == 0)
                    mat_imputation_d_u = matrix_sym_inverse(mat_imputation_omega_u, fail);
                // Now loop over elements of va (imputation_omega_e)
<% count = 0 %>\\
% for i in range(0, nresp1+nresp2):
% for j in range(0, i + 1):
                {
% if i < nresp2 and j < nresp2 and (context['imputation_resptype2_'+str(i+1)] != 'Normal'):
                    imputation_omega_u_accept[${count}] = 44;
% else:
                    auto ll_imputation_omega_u = [&](const SymMatrix &d_u) {
                        double ll = length(imputation_u0) * log(matrix_determinant(d_u)) / 2;
                        for (auto i = 0; i < ${nresp1+nresp2}; ++i) {
                            for (auto j = 0; j < ${nresp1+nresp2}; ++j) {
                                ll -= imputation_sbu(i, j) * d_u(i, j) / 2;
                            }
                        }
                        return ll;
                    };

                    double llcurr = ll_imputation_omega_u(mat_imputation_d_u);

                    double old = mat_imputation_omega_u(${i}, ${j});
                    mat_imputation_omega_u(${i}, ${j}) = dnorm(mat_imputation_omega_u(${i}, ${j}), 1.0/sqrt(imputation_omega_u_sd[${count}]));
                    SymMatrix temp_imputation_d_u = matrix_cholesky(mat_imputation_omega_u, fail);
                    if (fail == true) {
                        mat_imputation_omega_u(${i}, ${j}) = old;
                    } else {
                        temp_imputation_d_u = matrix_cholinverse(temp_imputation_d_u);
                        double llnew = ll_imputation_omega_u(temp_imputation_d_u);
                        bool change = true;
                        if(isFinite(llnew) == 0) {
                            change = false;
                        } else if (llnew > llcurr) {
                            ;
                        } else {
                            if (random01() > exp(llnew - llcurr)) {
                                change = false;
                            }
                        }
                        if (change == false) {
                            mat_imputation_omega_u(${i}, ${j}) = old;
                        } else {
                            imputation_omega_u_accept[${count}]++;
                            imputation_d_u_accept[${count}]++;
                            mat_imputation_d_u = temp_imputation_d_u;
                        }
                    }
% endif
<% count += 1 %>\\
                }
% endfor
% endfor

% endif
% endif

% if imputation_L1resp:
                // Step for y missing
                std::vector<double> xrow(mat_moi_x.cols());
% if moi_L2:
                std::vector<double> xrow2(mat_moi_x2.cols());
% endif
                auto ll_missing_y = [&](auto r) {
                    RectMatrix imputation_e(${nresp1}, 1);
% for j in range(0, nresp1):
                    imputation_e(${j}, 0) = mat_imputation_ymiss(r, ${j});
                    for (auto i = 0; i < mat_imputation_x${j + 1}.cols(); ++i) {
                        imputation_e(${j}, 0) -= mat_imputation_beta(betamap[${j}] + i, 0) * mat_imputation_x${j + 1}(r, i);
                    }
% if imputation_L2:
                    imputation_e(${j}, 0) -= mat_imputation_u(${L2ID}[r], ${j});
% endif
% endfor

                    SymMatrix imputation_sb1 = imputation_e*imputation_e.T();

                    double ll = 0.0;
                    for (auto i = 0; i < imputation_sb1.rows(); ++i) {
                        for (auto j = 0; j < imputation_sb1.cols(); ++j) {
                            ll -= imputation_sb1(i, j) * mat_imputation_d_e(i, j) / 2;
                        }
                    }

                    double xb = 0;
                    for (auto i = 0; i < mat_moi_x.cols(); i++) {
                        xb += mat_moi_x(r, i) * mat_moi_beta(i, 0);
                    }
% if moi_L2:
                    for (auto i = 0; i < ${len(moi_x2vars)}; ++i) {
                        xb += mat_moi_x2(r, i) * mat_moi_u(${L2ID}[r], i);
                    }
% endif

% if moi_D == 'Normal':
                    ll -= pow(mat_moi_y(r, 0) - xb, 2) * 0.5 * moi_tau;
% endif
% if moi_D == 'Binomial':
% if moi_link == 'logit':
                    double temp = -log1p(exp(-xb));
                    ll += (mat_moi_n(r, 0) - mat_moi_y(r, 0)) * (-xb + temp) + mat_moi_y(r, 0) * temp;
% endif
% if moi_link == 'probit':
                    double temp = normprob(-xb);
                    ll += (mat_moi_n(r, 0) - mat_moi_y(r, 0)) * log(temp) + mat_moi_y(r, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
                    double temp = -exp(xb);
                    ll += (mat_moi_n(r, 0) - mat_moi_y(r, 0)) * temp + mat_moi_y(r, 0) * log(1.0 - exp(temp));
% endif
% endif
% if moi_D == 'Poisson':
% if moi_offset:
                    xb += mat_moi_y(r, 0);
% endif
                    ll += mat_moi_y(r, 0) * xb - exp(xb);
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
                    xb += mat_moi_n(r, 0);
% endif
                    ll += mat_moi_y(r, 0) * xb - exp(xb) * mat_moi_rho(r, 0);
% endif
                    return ll;
                };

% for i in range(0, len(imputation_y)):
                for (int j = 0; j < mat_imputation_ymiss.rows(); j++) {
% if context['imputation_resptype'+str(i+1)] == 'Normal':
                    if (mat_imputation_y(j, ${i}) <= -9.999e29) {
% endif

// Starts here
                        bool change = true;
                        double llcurr = ll_missing_y(j);

                        for (int a = 0; a < mat_moi_x.cols(); a++) {
                            xrow[a] = mat_moi_x(j, a);
                        }
% if moi_L2:
                        for (int a = 0; a < mat_moi_x2.cols(); a++) {
                            xrow2[a] = mat_moi_x2(j, a);
                        }
% endif

                        double old = mat_imputation_ymiss(j, ${i});
% if context['imputation_resptype'+str(i+1)] == 'Normal':
                        mat_imputation_ymiss(j, ${i}) = dnormsd(mat_imputation_ymiss(j, ${i}), miss${imputation_y[i]}_sd[j]);
% endif
% if context['imputation_resptype'+str(i+1)] == 'Binary':
                        mat_imputation_ymiss(j, ${i}) = dnormsd(mat_imputation_ymiss(j, ${i}), misb${imputation_y[i]}_sd[j]);
                        if (${imputation_y[i]}[j] == 0 && mat_imputation_ymiss(j, ${i}) > 0) {
                            change = false;
                        } else if (${imputation_y[i]}[j] == 1 && mat_imputation_ymiss(j, ${i}) < 0) {
                            change = false;
                        } else {
                            impb${imputation_y[i]}[j] = (mat_imputation_ymiss(j, ${i}) > 0);
                        }
% endif

                        recalc_x(j);
% if moi_L2:
                        recalc_x2(j);
% endif

                        double llnew = ll_missing_y(j);

                        if (isFinite(llnew) == 0) {
                            change = false;
                        } else if (llnew > llcurr) {
                            ;
                        } else {
                            if(random01() > exp(llnew-llcurr)) {
                                change = false;
                            }
                        }
                        if (change == false) {
                            mat_imputation_ymiss(j, ${i}) = old;
                            for (int a = 0; a < mat_moi_x.cols(); a++) {
                                if (const_x[a] == false) {
                                    mat_moi_x(j, a) = xrow[a];
                                }
                            }
% if moi_L2:
                            for (int a = 0; a < mat_moi_x2.cols(); a++) {
                                if (const_x2[a] == false) {
                                    mat_moi_x2(j, a) = xrow2[a];
                                }
                            }
% endif
% if context['imputation_resptype'+str(i+1)] == 'Binary':
                            if (${imputation_y[i]}[j] <= -9.999e29) {
                                impb${imputation_y[i]}[j] = (mat_imputation_ymiss(j, ${i}) > 0);
                            }
% endif
                        } else {
% if context['imputation_resptype'+str(i+1)] == 'Normal':
                            miss${imputation_y[i]}_accept[j]++;
% endif
% if context['imputation_resptype'+str(i+1)] == 'Binary':
                            misb${imputation_y[i]}_accept[j]++;
                            impb${imputation_y[i]}_accept[j]++;
% endif
                        }
% if context['imputation_resptype'+str(i+1)] == 'Normal':
                    } else {
                            miss${imputation_y[i]}_accept[j] = 44;
                    }
% endif
                }
% endfor

% endif

% if imputation_L2resp:

                // Step for y2 missing
                //std::vector<double> xrow(mat_moi_x.cols());
% if moi_L2:
                //std::vector<double> xrow2(mat_moi_x2.cols());
% endif
                auto ll_missing_y2 = [&](auto r) {
                    RectMatrix imputation_e(${nresp2 + nresp1}, 1);
% for j in range(0, nresp2):
                    imputation_e(${j}, 0) = mat_imputation_ymiss2(r, ${j});
                    for (auto i = 0; i < mat_imputation_x2_${j + 1}.cols(); ++i) {
                        imputation_e(${j}, 0) -= mat_imputation_beta2(betamap2[${j}] + i, 0) * mat_imputation_x2_${j + 1}(r, i);
                    }
% endfor
% for j in range(0, nresp1):
                    imputation_e(${j + nresp2}, 0) = mat_imputation_u(r, ${j});
% endfor
                    SymMatrix imputation_sb2 = imputation_e*imputation_e.T();

                    double ll = 0.0;
                    for (auto i = 0; i < imputation_sb2.rows(); ++i) {
                        for (auto j = 0; j < imputation_sb2.cols(); j++) {
                            ll -= imputation_sb2(i, j) * mat_imputation_d_u(i, j) / 2;
                        }
                    }

                    for (auto id_k = 0; id_k < idmap_${L2ID}[r].size(); ++id_k) {
                        auto k = idmap_${L2ID}[r][id_k];

                        double xb = 0;
                        for (auto i = 0; i < mat_moi_x.cols(); ++i) {
                            xb += mat_moi_x(k, i) * mat_moi_beta(i, 0);
                        }
% if moi_L2:
                        for (auto i = 0; i < ${len(moi_x2vars)}; ++i) {
                            xb += mat_moi_x2(k, i) * mat_moi_u(r, i);
                        }
% endif

% if moi_D == 'Normal':
                        ll -= pow(mat_moi_y(k, 0) - xb, 2) * 0.5 * moi_tau;
% endif
% if moi_D == 'Binomial':
% if moi_link == 'logit':
                        double temp = -log1p(exp(-xb));
                        ll += (mat_moi_n(k, 0) - mat_moi_y(k, 0)) * (-xb + temp) + mat_moi_y(k, 0) * temp;
% endif
% if moi_link == 'probit':
                        double temp = normprob(-xb);
                        ll += (mat_moi_n(k, 0) - mat_moi_y(k, 0)) * log(temp) + mat_moi_y(k, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
                        double temp = -exp(xb);
                        ll += (mat_moi_n(k, 0) - mat_moi_y(k, 0)) * temp + mat_moi_y(k, 0) * log(1.0 - exp(temp));
% endif
% endif
% if moi_D == 'Poisson':
% if moi_offset:
                        xb += mat_moi_n(k, 0);
% endif
                        ll += mat_moi_y(k, 0) * xb - exp(xb);
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
                        xb += mat_moi_n(k, 0);
% endif
                        ll += mat_moi_y(k, 0) * xb - exp(xb) * mat_moi_rho(k, 0);
% endif
                    }
                    return ll;
                };

% for i in range(0, len(imputation_y2)):
                for (int j = 0; j < mat_imputation_ymiss2.rows(); j++) {
% if context['imputation_resptype2_' + str(i + 1)] == 'Normal':
                    if (mat_imputation_y2(j, ${i}) <= -9.999e29) {
% endif

// Starts here
                        bool change = true;
                        double llcurr = ll_missing_y2(j);
                        for (int a = 0; a < mat_moi_x.cols(); a++) {
                            xrow[a] = mat_moi_x(j, a);
                        }
% if moi_L2:
                        for (int a = 0; a < mat_moi_x2.cols(); a++) {
                            xrow2[a] = mat_moi_x2(j, a);
                        }
% endif

                        double old = mat_imputation_ymiss2(j, ${i});
% if context['imputation_resptype2_'+str(i+1)] == 'Normal':
                        mat_imputation_ymiss2(j, ${i}) = dnormsd(mat_imputation_ymiss2(j, ${i}), miss${imputation_L2Data}_${imputation_y2[i]}_sd[j]);
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                            int k = idmap_${L2ID}[j][id_k];
                            miss${imputation_y2[i]}[k] = mat_imputation_ymiss2(j, ${i});
                        }
% endif
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
                        mat_imputation_ymiss2(j, ${i}) = dnormsd(mat_imputation_ymiss2(j, ${i}), misb${imputation_L2Data}_${imputation_y2[i]}_sd[j]);
                        if (${imputation_L2Data}_${imputation_y2[i]}[j] == 0 && mat_imputation_ymiss2(j, ${i}) > 0) {
                            change = false;
                        } else if (${imputation_L2Data}_${imputation_y2[i]}[j] == 1 && mat_imputation_ymiss2(j, ${i}) < 0){
                            change = false;
                        } else {
                            impb${imputation_L2Data}_${imputation_y2[i]}[j] = (mat_imputation_ymiss2(j, ${i}) > 0);
                            for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                                int k = idmap_${L2ID}[j][id_k];
                                impb${imputation_y2[i]}[k] = impb${imputation_L2Data}_${imputation_y2[i]}[j];
                            }
                        }
% endif

                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                            int k = idmap_${L2ID}[j][id_k];
                            recalc_x(k);
% if moi_L2:
                            recalc_x2(k);
% endif
                        }

                        double llnew = ll_missing_y2(j);

                        if (isFinite(llnew) == 0) {
                            change = false;
                        } else if (llnew > llcurr) {
                            ;
                        } else {
                            if (random01() > exp(llnew-llcurr)) {
                                change = false;
                            }
                        }
                        if (change == false) {
                            mat_imputation_ymiss2(j, ${i}) = old;
                            for (int a = 0; a < mat_moi_x.cols(); a++) {
                                if (const_x[a] == false) {
                                    mat_moi_x(j, a) = xrow[a];
                                }
                            }
% if moi_L2:
                            for (int a = 0; a < mat_moi_x2.cols(); a++) {
                                if (const_x2[a] == false) {
                                    mat_moi_x2(j, a) = xrow2[a];
                                }
                            }
% endif

% if context['imputation_resptype2_'+str(i+1)] == 'Normal':
                            for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                                int k = idmap_${L2ID}[j][id_k];
                                miss${imputation_y2[i]}[k] = mat_imputation_ymiss2(j, ${i});
                            }
% endif
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
                            if (${imputation_L2Data}_${imputation_y2[i]}[j] <= -9.999e29) {
                                impb${imputation_L2Data}_${imputation_y2[i]}[j] = (mat_imputation_ymiss2(j, ${i}) > 0);
                                for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                                    int k = idmap_${L2ID}[j][id_k];
                                    impb${imputation_y2[i]}[k] = impb${imputation_L2Data}_${imputation_y2[i]}[j];
                                }
                            }
% endif
                        } else {
% if context['imputation_resptype2_'+str(i+1)] == 'Normal':
                            miss${imputation_L2Data}_${imputation_y2[i]}_accept[j]++;
                            for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                                int k = idmap_${L2ID}[j][id_k];
                                miss${imputation_y2[i]}_accept[k]++;
                            }
% endif
% if context['imputation_resptype2_'+str(i+1)] == 'Binary':
                            misb${imputation_L2Data}_${imputation_y2[i]}_accept[j]++;
                            impb${imputation_L2Data}_${imputation_y2[i]}_accept[j]++;
                            for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                                int k = idmap_${L2ID}[j][id_k];
                                impb${imputation_y2[i]}_accept[k]++;
                            }
% endif
                        }
% if context['imputation_resptype2_' + str(i + 1)] == 'Normal':
                    } else {
                        miss${imputation_L2Data}_${imputation_y2[i]}_accept[j] = 44;
                        for (int id_k = 0; id_k < idmapsizes_${L2ID}[j]; ++id_k) {
                            int k = idmap_${L2ID}[j][id_k];
                            miss${imputation_y2[i]}_accept[k] = 44;
                        }
                    }
% endif
                }
% endfor
% endif
            }
         '''

    diccode_temp = '''
        std::vector<double*> tmp_moi_x;
% for i in range(0, len(moi_xvars)):
% if imputation_L1resp and moi_xvars[i] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(moi_xvars[i])+1)] == 'Normal':
                tmp_moi_x.push_back(miss${moi_xvars[i]});
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(moi_xvars[i])+1)] == 'Binary':
                tmp_moi_x.push_back(impb${moi_xvars[i]});
% endif
% else:
% if moi_xvars[i] in moi_x_src:
                tmp_moi_x.push_back(const_cast<double *>(${moi_xvars[i]}));
% else:
                std::vector<double> moi_x${i}(length(${moi_xvars[0]}));
                for (int i = 0; i < length(${moi_xvars[0]}); i++) {
                    moi_x${i}[i] = \\
<% xterms = moi_xvars[i].split('*') %>\\
% for j in range(0, len(xterms)):
% if j != 0:
 *\\
% endif
<% xpoly = xterms[j].split('^') %>\\
% if len(xpoly) > 1:
 pow(\\
% endif
% if imputation_L1resp and xpoly[0] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(xpoly[0])+1)] == 'Normal':
 miss${xpoly[0]}[i]\\
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(xpoly[0])+1)] == 'Binary':
 impb${xpoly[0]}[i]\\
% endif
% elif imputation_L2 and imputation_L2resp and xpoly[0] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(xpoly[0])+1)] == 'Normal':
 miss${xpoly[0]}[i]\\
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(xpoly[0])+1)] == 'Binary':
 impb${xpoly[0]}[i]\\
% endif
% else:
 ${xpoly[0]}[i]\\
% endif
% if len(xpoly) > 1:
, ${xpoly[1]})\\
% endif
% endfor
;
                }
                tmp_moi_x.push_back(const_cast<double *>(&moi_x${i}[0]));
% endif
% endif
% endfor
                RectMatrix mat_moi_x(tmp_moi_x, length(${moi_x_src[0]}));



% if moi_L2:
                std::vector<double*> tmp_moi_x2;
% for i in range(0, len(moi_x2vars)):
% if imputation_L1resp and moi_x2vars[i] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(moi_x2vars[i])+1)] == 'Normal':
                tmp_moi_x2.push_back(miss${moi_x2vars[i]});
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(moi_x2vars[i])+1)] == 'Binary':
                for (int i = 0; i < length(${moi_x2vars[0]}); ++i) {
                    if (${moi_xvars[i]}[i] <= -9.999e29) {
                        impb${moi_x2vars[i]}[i] = (misb${moi_x2vars[i]}[i] > 0);
                    }
                }
                tmp_moi_x2.push_back(impb${moi_x2vars[i]});
% endif
% else:
% if moi_x2vars[i] in moi_x2_src:
                tmp_moi_x2.push_back(const_cast<double *>(${moi_x2_src[list(moi_x2_src).index(moi_x2vars[i])]}));
% else:
                std::vector<double> moi_x2${i}(length(${moi_x2vars[0]}));
                for (int i = 0; i < length(${moi_x2_src[0]}); i++) {
                    moi_x2${i}[i] = \\
<% x2terms = moi_x2vars[i].split('*') %>\\
% for j in range(0, len(x2terms)):
% if j != 0:
 *\\
% endif
<% x2poly = x2terms[j].split('^') %>\\
% if len(x2poly) > 1:
 pow(\\
% endif
% if imputation_L1resp and x2poly[0] in imputation_y:
% if context['imputation_resptype'+str(list(imputation_y).index(x2poly[0])+1)] == 'Normal':
 miss${x2poly[0]}[i]\\
% endif
% if context['imputation_resptype'+str(list(imputation_y).index(x2poly[0])+1)] == 'Binary':
 impb${x2poly[0]}[i]\\
% endif
% elif imputation_L2 and imputation_L2resp and x2poly[0] in imputation_y2:
% if context['imputation_resptype2_'+str(list(imputation_y2).index(x2poly[0])+1)] == 'Normal':
 miss${x2poly[0]}[i]\\
% endif
% if context['imputation_resptype2_'+str(list(imputation_y2).index(x2poly[0])+1)] == 'Binary':
 impb${x2poly[0]}[i]\\
% endif
% else:
 ${x2poly[0]}[i]\\
% endif
% if len(x2poly) > 1:
, ${x2poly[1]})\\
% endif
% endfor
;
                }
                tmp_moi_x2.push_back(const_cast<double *>(&moi_x2${i}[0]));
% endif
% endif
% endfor
                RectMatrix mat_moi_x2(tmp_moi_x2, length(${moi_x2_src[0]}));
% endif

% if moi_L2:
                std::vector<double*> tmp_moi_u;
% for i in range(0, len(moi_x2vars)):
                tmp_uint.push_back(const_cast<double *>(moi_u${i}));
% endfor
                RectMatrix mat_moi_u(tmp_moi_u, length(moi_u0));
% endif


        RectMatrix moi_ydiff(mat_moi_y);
% if moi_D == 'Normal':
% if moi_L2:
        for (int i = 0; i < imputation_ydiff.rows(); i++) {
            for (int m = 0; m < ${len(moi_x2vars)}; ++m) {
                moi_ydiff(i, 0) -= mat_moi_x2(i, m) * mat_moi_u(${L2ID}[i], m);
            }
        }
% endif
        RectMatrix yxb = moi_ydiff - mat_moi_x * mat_moi_beta;
        deviance = (moi_tau*(yxb.T() * yxb)(0,0)) + (log(3.14159265)-log(moi_tau))*length(${moi_y}) + (2.0*0.346573590279973*length(${moi_y})); // Last term is log(sqrt(2))
% else:

        deviance = 0.0;
        for (int j = 0; j < mat_moi_x.rows(); ++j) {
            double xb = 0;
            for (int k = 0; k < mat_moi_x.cols(); ++k) {
                xb += mat_moi_x(j, k) * mat_moi_beta(k, 0);
            }
% if moi_L2:
            for (int m = 0; m < ${len(moi_x2vars)}; ++m) {
                xb += mat_moi_x2(j, m) * mat_moi_u(${L2ID}[j], m);
            }
% endif
% if moi_D == 'Binomial':
% if moi_link == 'logit':
            double temp = -log1p(exp(-xb));
            deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * (-xb + temp) + mat_moi_y(j, 0) * temp;
% endif
% if moi_link == 'probit':
            double temp = normprob(-xb);
            deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * log(temp) + mat_moi_y(j, 0) * log(1.0 - temp);
% endif
% if moi_link == 'cloglog':
            double temp = -exp(xb);
            deviance += (mat_moi_n(j, 0) - mat_moi_y(j, 0)) * temp + mat_moi_y(j, 0) * log(1.0 - exp(temp));
% endif
            deviance += logfact(mat_moi_n(j, 0)) - logfact(mat_moi_y(j, 0)) - logfact(mat_moi_n(j, 0) - mat_moi_y(j, 0));
% endif
% if moi_D == 'Poisson':
% if moi_offset:
            xb += mat_moi_n(j, 0);
% endif
            deviance += mat_moi_y(j, 0) * xb - exp(xb);
            deviance += logfact(mat_moi_y(j, 0));
% endif
% if moi_D == '-ve Binomial':
% if moi_offset:
            xb += mat_moi_n(j, 0);
% endif
            deviance += mat_moi_y(j, 0) * xb - exp(xb) * mat_moi_rho(j, 0);
            deviance -= logfact(mat_moi_y(j, 0));
% endif
        }
        deviance *= -2;
% endif
'''

    preparedata = '''
mydata = data['datafile']
if imputation_L1resp:
    for i in range(0, len(imputation_y)):
        if (context['imputation_resptype'+str(i+1)] == 'Normal'):
            context['miss'+imputation_y[i]].ncols = -1*mydata.nobs
        if (context['imputation_resptype'+str(i+1)] == 'Binary'):
            context['impb'+imputation_y[i]].ncols = -1*mydata.nobs
            context['misb'+imputation_y[i]].ncols = -1*mydata.nobs

if imputation_L2 and imputation_L2resp:
    mydata2 = data[str(imputation_L2Data)]
    for i in range(0, len(imputation_y2)):
        if context['imputation_resptype2_'+str(i+1)] == 'Normal':
            context['miss'+imputation_y2[i]].ncols = -1*mydata.nobs
        if context['imputation_resptype2_'+str(i+1)] == 'Binary':
            context['impb'+imputation_y2[i]].ncols = -1*mydata.nobs

    # Prevent variable being imputed at level-2 being listwise deleted from level-1 by replacing missing values with zero
    for xvar in moi_x_src:
        if xvar in imputation_y2:
            mydata.variables[xvar]['data'][:] = mydata.variables[xvar]['data'].filled(0)

    for xvar in moi_x2_src:
        if xvar in imputation_y2:
            mydata.variables[xvar]['data'][:] = mydata.variables[xvar]['data'].filled(0)
'''
