# Copyright (c) 2013, University of Bristol and University of Southampton.

## Template dependencies:
# 1LevelMod.py
# 1LevelMVMixedResponsecc.py
# 1LevelMVNormal.py
# CompleteCases.py
# Merge.py
# NLevelMod.py
# NLevelMVNormal.py
# NLevelRS.py
# RespNLevelMVMixedResponsemvu.py
# Take.py

from EStat.Templating import Template

class TemplateNLevelImpute(Template):
    'A (super)template for N-level data that allows for missing values in explanatory or response variables, and can handle normal or categorical variables. After the user separately specifies imputation and model of interest (MOI), multiply-imputed datasets are produced, and the MOI fitted to each (via a variety of other templates); these are combined using Rubin\'s rules, and the results for the final MOI fit are then returned (together with results from a complete case analysis, for comparison). If your computer has multiple processors, these will be used in parallel for the imputation models.'

    __version__ = '1.0.0'

    tags = [ 'Multiple imputation', 'Missing data', '1-Level', 'N-Level', 'Model', 'Multivariate response', 'Normal', 'Ordered multinomial', 'Random slopes', 'Unordered multinomial' ]
    engines = ['Python_script']

    inputs = '''
higherlevs = Boolean('Do either the model of interest (MOI) and/or imputation have higher classifications?')

###specify MOI:
if not higherlevs:
    MOIdist = Text('MOI specification: what distribution would you like to use to model the response variable(s) in your MOI?', ['Normal', 'Binary', 'Poisson', 'Multivariate Normal'])
    if MOIdist == 'Normal' or MOIdist == 'Binary' or MOIdist == 'Poisson':
        templatename = Text(value = '1LevelMod')
    else:
        templatename = Text(value = '1LevelMVNormal')

def alpha_counter(i):
    from string import ascii_uppercase
    a = i % 26
    b = i / 26
    if b > 0:
        return alpha_counter(b-1) + ascii_uppercase[a]
    else:
        return ascii_uppercase[a]

if higherlevs:
    NumLevs = Integer('Number of higher level classifications: ', help='Select the variable(s) containing the identifying code for the higher-level classification(s) for which you would like random effects to be added to your model.')
    for i in range(0, int(NumLevs)):
        #context['C' + str(i + 1)] = IDVector('Higher level classification ' + str(i + 1) + ': ')
        context['C' + str(i + 1)] = IDVector('Higher level classification ' + alpha_counter(i) + ': ')

    if int(NumLevs) > 0:
        MOIlevel = Text('MOI specification: in your MOI, do you want higher level classifications?', ['no, just 1 level', 'yes'], help = "If you answer 'yes', the higher level classifications you nominated earlier will be used in the random part of the MOI.")
        MOIRespHighLev = Boolean('In your MOI, are there responses at any higher level classifications?')
        if MOIRespHighLev:
            MOINhigherlevels = Integer ('Number of higher level classifications at which there are responses')
            endlev = MOINhigherlevels + 2
## NOTE would it be best if user presented with classification IDs they previously nominated, so can choose from these?
            for lev in range(2, endlev):
                context['MOIL'+str(lev)+'ID'] = IDVector('Classification ' + str(lev-1)+ ' ID: ')
            templatename = Text(value = 'RespNLevelMVMixedResponsemvu')
        else:
            MOIdist = Text('What distribution would you like to use to model the response variable(s) in your MOI?', ['Normal', 'Binary', 'Poisson', 'Multivariate Normal'])
            if MOIdist == 'Normal' or MOIdist == 'Binary' or MOIdist == 'Poisson':
                if MOIlevel == 'yes':
                    MOIslope = Boolean('In your MOI, do you want to fit random slope/coefficient(s) for any of your higher level classifications?')
                    if MOIslope:
                        templatename = Text(value = 'NLevelRS')
                    else:
                        templatename = Text(value = 'NLevelMod')
                else:
                        templatename = Text(value = '1LevelMod')                
            else:
                if MOIlevel == 'yes':
                    templatename = Text(value = 'NLevelMVNormal')
                else:
                    templatename = Text(value = '1LevelMVNormal')

if templatename == 'NLevelMVNormal' or templatename == '1LevelMVNormal' or templatename == 'RespNLevelMVMixedResponsemvu':
    y = DataMatrix('In your MOI, what are the response variables (at level 1)? ')

    if templatename == 'RespNLevelMVMixedResponsemvu':
        for i in range(0, len(y)):
            context['MOIbin1_'+str(i+1)] = Text('Response Type for response ' + y[i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
            if context['MOIbin1_'+str(i+1)] == 'Ordered':
                context['MOIncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
            if context['MOIbin1_'+str(i+1)] == 'Unordered':
                context['MOIncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
                for j in range(0, context['MOIncats'+y[i]]-1):
                    context['MOIx1_'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
            else:
                context['MOIx1_'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)   

        for lev in range(0, int(MOINhigherlevels)):
## NOTE should this question be: what are your classification blah responses? As user already indicated the classifications for which there are higher-level responses, so just need to match them up now...
            context['MOIanyL'+str(lev+2)+'resp'] = Boolean('Are there any classification ' + str(lev+1) + ' (' + context['MOIL'+str(lev+2)+'ID'] + ') responses:')
            if context['MOIanyL'+str(lev+2)+'resp']:
                context['y'+str(lev+2)] = DataMatrix('Responses at classification ' + str(lev+1) + ' ('+ context['MOIL'+str(lev+2)+'ID'] +'): ')
                for i in range(0, len(context['y'+str(lev+2)])):
                    context['MOIbin'+str(lev+2)+'_'+str(i+1)] = Text('Response Type for response ' + context['y' + str(lev+2)][i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
                    if context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Ordered':
                        context['MOIncats'+context['y'+str(lev+2)][i]] = Integer('Number of categories for ' + context['y'+str(lev+2)][i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
                    if context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Unordered':
                        context['MOIncats'+context['y'+str(lev+2)][i]] = Integer('Number of categories for ' + context['y'+str(lev+2)][i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
                        for j in range(0, context['MOIncats'+context['y'+str(lev+2)][i]]-1):
                            context['MOIx'+str(lev+2)+'_'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
                    else:
                        context['MOIx' +str(lev+2)+'_'+str(i+1)] = DataMatrix('Explanatory variables for response ' + context['y'+str(lev+2)][i] + ': ', allow_cat = True)
    else:
        for i in range(0, len(y)):
            context['x'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ' in your MOI: ', allow_cat = True)

    if len(y) > 1:
        priors = Text('Priors: ', ['Uniform', 'Wishart'])
        if priors == 'Wishart':
            R = List('R matrix: ')
            v = Integer('Degrees of Freedom:')
        if higherlevs:
            for i in range(0, int(NumLevs)):
                context['priors' + str(i+1)] = Text('Classification ' + context['C'+str(i+1)] + ' priors: ', ['Uniform', 'Wishart'])
                if context['priors' + str(i+1)] == 'Wishart':
                    context['R' + str(i+1)] = List('R matrix: ')
                    context['v' + str(i+1)] = Integer('Degrees of Freedom:')
    mv = Boolean('Use MVNormal update for beta?: ')
else:
    if MOIdist == 'Normal':
        D = Text(value = 'Normal')
    elif MOIdist == 'Binary':
        D = Text(value = 'Binomial')
    elif MOIdist == 'Poisson':
        D = Text(value = 'Poisson')
    y = DataVector('In your MOI, what is the response variable? ')
#    D = Text('Response distribution type: ', ['Normal', 'Binomial', 'Poisson'])
    if D == 'Binomial':
#        n = DataVector('What is the binomial denominator in your MOI? ')
        link = Text('In your MOI, what link function would you like to use? ', ['logit', 'probit', 'cloglog'])
    if D == 'Poisson':
        link = Text(value = 'ln')
        offset = Boolean('Is there an offset? ')
        if offset:
            n = DataVector('offset: ')
    x = DataMatrix('Explanatory variables for MOI: ', allow_cat = True)
if templatename == 'NLevelRS':
    for i in range(0, int(NumLevs)):
        context['x' + str(i+1)] = DataMatrix('In your MOI, which explanatory variable coefficients are random at classification ' + str(i+1) + ' (' + context['C'+str(i+1)]+')? ', allow_cat = True)
        context['priorsMOI' + str(i+1)] = Text('MOI Priors for random coefficients at classification ' + str(i+1) + ' (' + context['C'+str(i+1)]+'):', ['Uniform', 'Wishart'])
        if context['priorsMOI' + str(i+1)] == 'Wishart':
            context['RMOI' + str(i+1)] = List('R matrix: ')
            context['vMOI' + str(i+1)] = Integer('Degrees of Freedom:')

###specify imputation model:
###(changed so asks at a more logical pace in eBook)

yimp1 = DataMatrix('Now we will specify the imputation model: responses at level 1: ')
for i in range(0, len(yimp1)):
    context['bin1_'+str(i+1)] = Text('Response Type for response ' + yimp1[i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
for i in range(0, len(yimp1)):
    if(context['bin1_'+str(i+1)] == 'Ordered'):
       context['ncats'+yimp1[i]] = Integer('Number of categories (numbered 0,1..) for ' + yimp1[i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
    if(context['bin1_'+str(i+1)] == 'Unordered'):
       context['ncats'+yimp1[i]] = Integer('Number of categories (numbered 0,1..) for ' + yimp1[i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
    if(context['bin1_'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+yimp1[i]]-1):
            context['ximp1' + '_' + str(i) + '_' + str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ' in imputation model: ', allow_cat = True)
    else:
        context['ximp1' + '_' + str(i)] = DataMatrix('Explanatory variables for ' + yimp1[i] + ' in imputation model: ', allow_cat = True)

if higherlevs:
    IMlevel = Text('In your IM, do you want higher level classifications?', ['no, just 1 level', 'yes'], help = "If you answer 'yes', the higher level classifications you nominated earlier will be used in the random part of the IM.")
    want2 = Boolean('Are there any responses for higher level classifications in the imputation model?: ')
    for j in range(0, int(NumLevs)):
        if not want2:
            context['anyL'+str(j+2)+'resp'] = Boolean(value='No')
        else:
            context['anyL'+str(j+2)+'resp'] = Boolean('Are there any responses for classification ' + context['C' + str(j + 1)] + ':')
            if context['anyL'+str(j+2)+'resp']:
                context['yimp' + str(j+2)] = DataMatrix('Responses at classification ' + str(j+1) + ' (' + context['C'+str(j+1)]+') for imputation model: ')
                for i in range(0, len(context['yimp' + str(j+2)])):
                    context['bin' + str(j+2) + '_'+str(i+1)] = Text('Response Type for response ' + context['yimp' + str(j+2)][i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
                for i in range(0, len(context['yimp' + str(j+2)])):
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Ordered'):
                       context['ncats'+context['yimp' + str(j+2)][i]] = Integer('Number of categories (numbered 0,1..) for ' + context['yimp' + str(j+2)][i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Unordered'):
                       context['ncats'+context['yimp' + str(j+2)][i]] = Integer('Number of categories (numbered 0,1..) for ' + context['yimp' + str(j+2)][i] + ': ', help = "Your categories, as represented in your dataset, need to be numbered from zero, sequentially in steps of one (i.e. 0,1,2 if you had 3 categories).")
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Unordered'):
                        for k in range(0, context['ncats'+context['yimp' + str(j+2)][i]]-1):
                            context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)] = DataMatrix('Explanatory variables for category ' + str(k) + ' at classification ' + str(j+1) + ' (' + context['C'+str(j+1)]+') for imputation model: ', allow_cat = True)
                    else:
                        context['ximp' + str(j+2) + '_' + str(i)] = DataMatrix('Explanatory variables for response ' + context['yimp' + str(j+2)][i] + ' at classification ' + str(j+1) + ' (' + context['C'+str(j+1)]+') for imputation model: ', allow_cat = True)

                # Check response is really at level N
                classvar = GUIObject.interface.data[context['C' + str(j + 1)].data].variables[str(context['C' + str(j + 1)])]['data']
                (classvals, classindex, classinverse) = numpy.unique(classvar, return_index=True, return_inverse=True)
                for i in range(0, len(context['yimp' + str(j+2)])):
                    datavar = GUIObject.interface.data[context['yimp'+str(j+2)].data].variables[context['yimp'+str(j+2)][i]]['data']
                    if not numpy.all(datavar == datavar[classindex][classinverse]):
                        context['msg' + str(i)] = Feedback(str(context['yimp'+str(j+2)][i]) + ' is not a level ' + str(j+2) + ' variable')
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Unordered'):
                        for m in range(0, len(context['ximp' + str(j+2) + '_' + str(i) + '_' + str(j)])):
                            for k in range(0, context['ncats'+context['yimp' + str(j+2)][i]]-1):
                                datavar = GUIObject.interface.data[context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)].data].variables[context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)][i]]['data']
                                if not numpy.all(datavar == datavar[classindex][classinverse]):
                                    context['msg' + str(i) + '_' + str(m) + '_' + str(k)] = Feedback(str(context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)][i]) + ' is not a level ' + str(j+2) + ' variable')
                    else:
                        for m in range(0, len(context['ximp' + str(j+2) + '_' + str(i)])):
                            datavar = GUIObject.interface.data[context['ximp' + str(j+2) + '_' + str(i)].data].variables[context['ximp' + str(j+2) + '_' + str(i)][m]]['data']
                            if not numpy.all(datavar == datavar[classindex][classinverse]):
                                context['msg' + str(i) + '_' + str(m)] = Feedback(str(context['ximp'+str(j+2) + '_' + str(i)][j]) + ' is not a level ' + str(j+2) + ' variable')

###changed condmarg from Boolean to TextRadio so doesn't have default in eBook (led to problems of it being selected in error, by default, before user explicitly asked for input)

#condmarg = Boolean('Use conditional marginal algorithm: ')
condmarg = TextRadio('Use conditional marginal algorithm?', ['Yes', 'No'])

numimpute = Integer('Number of imputed data sets: ')

imputefirst = Integer('Number of iterations before first imputation: ', help="This is setting the interval between adaptation ending and the first imputation, i.e. it effectively sets the burn-in but is nevertheless included in diagnostics returned to aid model-checking.")
imputeevery = Integer('Number of iterations between subsequent imputations: ', help="If parallel chains are run, then only the 1st imputation is used; if not parallel chains (i.e. if fewer processors than imputed datasets requested), then this input question pertains to the interval at which iterations are subsequently sampled from the chain.")



burnin=Integer('Length of burnin for MOI: ')
iterations=Integer('Number of iterations for MOI: ')
'''

    pythonscript = '''
import numpy

import multiprocessing

import EStat
from EStat.Templating import *

#import sys
#sys.path.append(os.path.abspath(os.path.join('..', '..', 'apps', 'cmdtest')))

from cmdtest import RunStatJR
from EStat.DTAFile import DTAFile
from EStat.stats.utils.MCMCGraph import MCMCGraph
import datasets

print 'Datasets: ' + str(datasets.datasets.keys())

if templatename == 'NLevelMVNormal' or templatename == '1LevelMVNormal' or templatename == 'RespNLevelMVMixedResponsemvu':
    pass
else:
    if D == 'Binomial':
        datafile.addvariable("autodenom",data=numpy.ones(datafile.nobs))

datasets.datasets['workdata'] = datafile

nchain = multiprocessing.cpu_count()
nperchain = int(numimpute) / nchain
nextraimpute = int(numimpute) % nchain

imputes = []
for i in range(0, nchain):
    iters = []
    for j in range(0, nperchain):
        iters.append(j * int(imputeevery))
    if i < nextraimpute:
        iters.append(nperchain * int(imputeevery))
    if len(iters) > 0:
        imputes.append(iters)

moivars = set()
if templatename == 'NLevelMVNormal' or templatename == '1LevelMVNormal' or templatename == 'RespNLevelMVMixedResponsemvu':
    moivars.update(list(y))
    if templatename == 'RespNLevelMVMixedResponsemvu':
        for i in range(0, len(y)):
            if context['MOIbin1_'+str(i+1)] == 'Unordered':
                for j in range(0, context['MOIncats'+y[i]]-1):
                    moivars.update(list(context['MOIx1_'+str(i+1)+'_'+str(j)].orignames))
            else:
                moivars.update(list(context['MOIx1_'+str(i+1)].orignames))
        for lev in range(0, int(MOINhigherlevels)):
            if context['MOIanyL'+str(lev+2)+'resp']:
                for i in range(0, len(context['y'+str(lev+2)])):
                    if(context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Unordered'):
                        for j in range(0, context['MOIncats'+context['y'+str(lev+2)][i]]-1):
                            moivars.update(list(context['MOIx'+str(lev+2)+'_'+str(i+1)+'_'+str(j)].orignames))
                    else:
                        moivars.update(list(context['MOIx'+str(lev+2)+'_'+str(i+1)].orignames))
    else:
        for i in range(0, len(y)):
            moivars.update(list(context['x'+str(i+1)].orignames))
else:
    moivars.update([str(y)])
    moivars.update(list(x.orignames))
    if templatename == 'NLevelRS':
        moivars.update(list(x2.orignames))
    if D == 'Binomial':
        moivars.update(["autodenom"])
    if D == 'Poisson':
        moivars.update([str(n)])

if higherlevs:
    for j in range(0, int(NumLevs)):
        moivars.update([str(context['C' + str(j+1)])])

if MOIRespHighLev:
    for j in range(0, int(MOINhigherlevels)):
        moivars.update([str(context['MOIL' + str(j+2) + 'ID'])])

m = RunStatJR(template='CompleteCases', dataset='workdata', invars = {'vars': ','.join(moivars), 'outdata': 'CompleteData'}, estoptions = {'Engine': 'Python_script'})

if higherlevs and want2:
    for j in range(0, int(NumLevs)):
        if context['anyL'+str(j+2)+'resp']:
            classvars = list(context['yimp' + str(j+2)])
            for i in range(0, len(context['yimp' + str(j+2)])):
                if context['bin' + str(j+2) + '_'+str(i+1)] == 'Unordered':
                    for k in range(0, context['ncats'+context['yimp' + str(j+2)][i]]-1):
                        classvars += list(context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)])
                else:
                    classvars += list(context['ximp' + str(j+2) + '_' + str(i)])


            m = RunStatJR(template='Take', dataset='workdata', invars = {'LID': str(context['C' + str(j+1)]), 'incols': ','.join(classvars), 'outdata': '_C' + str(j+1) + 'Data'}, estoptions = {'Engine': 'Python_script'})

print 'Datasets: ' + str(datasets.datasets.keys())

inputs = {}


inputs['Nhigherlevels'] = str(NumLevs)

if higherlevs and (want2 or IMlevel):
    inputs['y1'] = ','.join(yimp1)
    inputs['storeresid'] = 'No'
    for j in range(0, int(NumLevs)):
        inputs['L' + str(j+2) + 'ID'] = str(context['C' + str(j+1)])
        inputs['anyL'+str(j+2)+'resp'] = str(context['anyL'+str(j+2)+'resp'])
    if want2:
        for j in range(0, int(NumLevs)):
            if context['anyL'+str(j+2)+'resp']:
                inputs['y' + str(j+2)] = ','.join(context['yimp' + str(j+2)])
                inputs['L' + str(j+2) + 'Data'] = '_C' + str(j+1) + 'Data'
else:
    inputs['y'] = ','.join(yimp1)

inputs['outdata'] = 'out'

inputs['condmarg'] = str(condmarg)

def inputlist(obj):
    varlist = []
    for i in range(0, len(obj.orignames)):
        if obj.orignames[i] in obj.categorical:
            varlist.append(obj.orignames[i] + ':cat') 
        else:
            varlist.append(obj.orignames[i])
    return varlist

if higherlevs and (want2 or IMlevel):
    for i in range(0, len(yimp1)):
        inputs['bin1_' + str(i + 1)] = str(context['bin1_'+str(i+1)])
        if(context['bin1_'+str(i+1)] == 'Ordered'):
            inputs['ncats'+yimp1[i]] = str(context['ncats'+yimp1[i]])
        if(context['bin1_'+str(i+1)] == 'Unordered'):
            inputs['ncats'+yimp1[i]] = str(context['ncats'+yimp1[i]])
            for j in range(0, context['ncats'+yimp1[i]] - 1):
                inputs['x1_'+str(i+1)+'_'+str(j)] = ','.join(inputlist(context['ximp1' + '_' + str(i) + '_' + str(j)])) #'_cons'
        else:
            inputs['x1_'+str(i+1)] = ','.join(inputlist(context['ximp1' + '_' + str(i)])) #'_cons'

    if want2:
        for j in range(0, int(NumLevs)):
            if context['anyL'+str(j+2)+'resp']:
                for i in range(0, len(context['yimp' + str(j+2)])):
                    inputs['bin' + str(j+2) + '_' + str(i + 1)] = str(context['bin' + str(j+2) + '_'+str(i+1)])
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Ordered'):
                        inputs['ncats'+context['yimp' + str(j+2)][i]] = str(context['ncats'+context['yimp' + str(j+2)][i]])
                    if(context['bin' + str(j+2) + '_'+str(i+1)] == 'Unordered'):
                        inputs['ncats'+context['yimp' + str(j+2)][i]] = str(context['ncats'+context['yimp' + str(j+2)][i]])
                        for k in range(0, context['ncats'+context['yimp' + str(j+2)][i]] - 1):
                            inputs['x' + str(j+2) + '_'+str(i+1)+'_'+str(k)] = ','.join(inputlist(context['ximp' + str(j+2) + '_' + str(i) + '_' + str(k)])) #'_cons'
                    else:
                        inputs['x' + str(j+2) + '_'+str(i+1)] = ','.join(inputlist(context['ximp' + str(j+2) + '_' + str(i)])) #'_cons'
else:
    for i in range(0, len(yimp1)):
        inputs['bin' + str(i + 1)] = str(context['bin1_'+str(i+1)])
        if(context['bin1_'+str(i+1)] == 'Ordered'):
            inputs['ncats'+yimp1[i]] = str(context['ncats'+yimp1[i]])
        if(context['bin1_'+str(i+1)] == 'Unordered'):
            inputs['ncats'+yimp1[i]] = str(context['ncats'+yimp1[i]])
            for j in range(0, context['ncats'+yimp1[i]] - 1):
                inputs['x'+str(i+1)+'_'+str(j)] = ','.join(inputlist(context['ximp1' + '_' + str(i) + '_' + str(j)])) #'_cons'
        else:
            inputs['x'+str(i+1)] = ','.join(inputlist(context['ximp1' + '_' + str(i)])) #'_cons'

omega_e1_diag = []

for i in range(0, len(yimp1)):
    if context['bin1_'+str(i+1)] == 'Normal':
        omega_e1_diag.append(str(datafile.variables[yimp1[i]]['data'].var()))
    elif context['bin1_'+str(i+1)] == 'Unordered':
        for j in range(0, int(context['ncats'+yimp1[i]]) - 1):
            omega_e1_diag.append('1')
    else:
        omega_e1_diag.append('1')

omega_e1_full = []
for i in range(0, len(omega_e1_diag)):
    for j in range(0, i+1):
        if i == j:
            omega_e1_full.append(omega_e1_diag[i])
        else:
            omega_e1_full.append('0')

if higherlevs and (want2 or IMlevel):
    inputs['omega_e1'] = ','.join(omega_e1_full)
else:
    inputs['omega_e'] = ','.join(omega_e1_full)

estinputs = {}
estinputs['Engine'] = 'CustomC'
estinputs['defaultsv'] = 'Yes'
estinputs['defaultalg'] = 'Yes'
estinputs['burnin'] = '0'
estinputs['seed'] = '1'
estinputs['thinning'] = '1'
estinputs['makepred'] = 'No'
# Note: impute iterations must be the same for all chains, so we are potentially doing more iterations than necessary
estinputs['nchains'] = str(len(imputes))
estinputs['iterations'] = str(int(imputefirst)+imputes[0][-1]+1)
estinputs['imputeiters'] = ','.join([str(int(imputefirst) + i) for i in imputes[0]])

if higherlevs and (want2 or IMlevel):
    m = RunStatJR(template='RespNLevelMVMixedResponsemvu', dataset='workdata', invars = inputs, estoptions = estinputs)
else:
    m = RunStatJR(template='1LevelMVMixedResponsecc', dataset='workdata', invars = inputs, estoptions = estinputs)

for k in m.eng.inputs.keys():
    outputs['Imputation_Model_' + k] = m.eng.inputs[k]

for k in m.eng.outputs.keys():
    outputs['Imputation_Model_' + k] = m.eng.outputs[k]

print("merging...")
if higherlevs and want2:
    for j in range(0, int(NumLevs)):
        if context['anyL'+str(j+2)+'resp']:
            for i in range(0, len(imputes)):
                for k in range(0, len(imputes[i])):
                    m = RunStatJR(template='Merge', dataset='impute_datafile_chain' + str(i) + '_iter' + str(int(imputefirst)+imputes[i][k]), invars = {'dataset2': 'impute__C' + str(j+1) +'Data_chain' + str(i) + '_iter' + str(int(imputefirst)+imputes[i][k]), 'outdata': 'impute_datafile_chain' + str(i) + '_iter' + str(int(imputefirst)+imputes[i][k]), 'idcol': str(context['C' + str(j+1)])}, estoptions = {'Engine': 'Python_script'})
                    for k in m.eng.outputs.keys():
                        outputs['Imputation_Model_' + k] = m.eng.outputs[k]
print("merged")

moiinputs = {}
if templatename == 'RespNLevelMVMixedResponsemvu':
    moiinputs['Nhigherlevels'] = str(MOINhigherlevels)
    for j in range(0, int(MOINhigherlevels)):
        moiinputs['L'+str(j+2)+'ID'] = str(context['MOIL'+str(j+2)+'ID'])
    moiinputs['storeresid'] = 'No'
    moiinputs['condmarg'] = 'Yes'
    moiinputs['y1'] = ",".join(y)
    moiinputs['imputeiters'] = '0'
    for i in range(0, len(y)):
        moiinputs['bin1_'+str(i+1)] = str(context['MOIbin1_'+str(i+1)])
        if context['MOIbin1_'+str(i+1)] == 'Ordered' or context['MOIbin1_'+str(i+1)] == 'Unordered':
            moiinputs['ncats'+y[i]] = str(context['MOIncats'+y[i]])
        if context['MOIbin1_'+str(i+1)] == 'Unordered':
            for j in range(0, context['MOIncats'+y[i]]-1):
                tmp = []
                for var in context['MOIx1_'+str(i+1)+'_'+str(j)].orignames:
                    if var in context['MOIx1_'+str(i+1)+'_'+str(j)].categorical:
                        tmp.append(var+':cat')
                    else:
                        tmp.append(var)
                moiinputs['x1_'+str(i+1)+'_'+str(j)] = ",".join(tmp)
        else:
            tmp = []
            for var in context['MOIx1_'+str(i+1)].orignames:
                if var in context['MOIx1_'+str(i+1)].categorical:
                    tmp.append(var+':cat')
                else:
                    tmp.append(var)
            moiinputs['x1_'+str(i+1)] = ",".join(tmp)
    for lev in range(0, int(MOINhigherlevels)):
        moiinputs['anyL'+str(lev+2)+'resp'] = str(context['MOIanyL'+str(lev+2)+'resp'])
        if context['MOIanyL'+str(lev+2)+'resp']:
            moiinputs['y'+str(lev+2)] = ",".join(context['y'+str(lev+2)])
            for i in range(0, len(context['y'+str(lev+2)])):
                moiinputs['bin'+str(lev+2)+'_'+str(i+1)] = str(context['MOIbin'+str(lev+2)+'_'+str(i+1)])
                if context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Ordered' or context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Unordered':
                    moiinputs['ncats'+context['y'+str(lev+2)][i]] = str(context['MOIncats'+context['y'+str(lev+2)][i]])
                if context['MOIbin'+str(lev+2)+'_'+str(i+1)] == 'Unordered':
                    for j in range(0, context['MOIncats'+context['y'+str(lev+2)][i]]-1):
                        tmp = []
                        for var in context['MOIx'+str(lev+2)+'_'+str(i+1)+'_'+str(j)].orignames:
                            if var in moiinputs['x'+str(lev+2)+'_'+str(i+1)+'_'+str(j)].categorical:
                                tmp.append(var+':cat')
                            else:
                                tmp.append(var)
                        moiinputs['x'+str(lev+2)+'_'+str(i+1)+'_'+str(j)] = ",".join(tmp)
                else:
                    tmp = []
                    for var in context['MOIx' +str(lev+2)+'_'+str(i+1)].orignames:
                        if var in context['MOIx' +str(lev+2)+'_'+str(i+1)].categorical:
                            tmp.append(var+':cat')
                        else:
                            tmp.append(var)
                    moiinputs['x'+str(lev+2)+'_'+str(i+1)] = ",".join(tmp)

    MOIomega_e1_diag = []

    for i in range(0, len(y)):
        if context['MOIbin1_'+str(i+1)] == 'Normal':
            MOIomega_e1_diag.append(str(datafile.variables[y[i]]['data'].var()))
        elif context['MOIbin1_'+str(i+1)] == 'Unordered':
            for j in range(0, int(context['MOIncats'+y[i]]) - 1):
                MOIomega_e1_diag.append('1')
        else:
            MOIomega_e1_diag.append('1')

    MOIomega_e1_full = []
    for i in range(0, len(MOIomega_e1_diag)):
        for j in range(0, i+1):
            if i == j:
                MOIomega_e1_full.append(MOIomega_e1_diag[i])
            else:
                MOIomega_e1_full.append('0')

    moiinputs['omega_e1'] = ','.join(MOIomega_e1_full)

    for j in range(0, int(MOINhigherlevels)):
        if context['MOIanyL'+str(j+2)+'resp']:
            classvars = list(context['y' + str(j+2)])
            for i in range(0, len(context['y' + str(j+2)])):
                if context['MOIbin' + str(j+2) + '_'+str(i+1)] == 'Unordered':
                     for k in range(0, context['MOIncats'+context['y' + str(j+2)][i]]-1):
                        classvars += list(context['MOIx' + str(j+2) + '_' + str(i+1) + '_' + str(k)])
                else:
                    classvars += list(context['MOIx' + str(j+2) + '_' + str(i+1)])

            m = RunStatJR(template='Take', dataset='workdata', invars = {'LID': str(context['MOIL' + str(j+2) + 'ID']), 'incols': ','.join(classvars), 'outdata': '_MOIL' + str(j+2) + 'Data'}, estoptions = {'Engine': 'Python_script'})
            for i in range(0, len(imputes)):
                for j in range(0, len(imputes[i])):
                    m = RunStatJR(template='Take', dataset='impute_datafile_chain' + str(i) + '_iter' + str(imputes[i][j]), invars = {'LID': str(context['MOIL' + str(j+2) + 'ID']), 'incols': ','.join(classvars), 'outdata': '_MOIL' + str(j+2) + 'Data' + str(i) + '_iter' + str(imputes[i][j])}, estoptions = {'Engine': 'Python_script'})
    print 'Datasets: ' + str(datasets.datasets.keys())

else:
    if higherlevs:
        moiinputs['NumLevs'] = str(NumLevs)
        for j in range(0, int(NumLevs)):
            moiinputs['C' + str(j+1)] = str(context['C' + str(j+1)])
        moiinputs['storeresid'] = 'No'
    if templatename == 'NLevelMVNormal' or templatename == '1LevelMVNormal':
        moiinputs['y'] = ",".join(y)
        for i in range(0, len(y)):
            moiinputs['x'+str(i+1)] = ",".join(context['x'+str(i+1)])
        if len(y) > 1:
            moiinputs['priors'] = str(priors)
            if priors == 'Wishart':
                moiinputs['R'] = ",".join(R)
                moiinputs['v'] = str(v)
            if higherlevs:
                for j in range(0, int(NumLevs)):
                    moiinputs['priors' + str(j+2)] = str(context['priors' + str(j)])
                    if context['priors' + str(j)] == 'Wishart':
                        moiinputs['R' + str(j)] = ",".join(context['R' + str(j+2)])
                        moiinputs['v' + str(j)] = str(context['v' + str(j+2)])
        moiinputs['mv'] = str(mv) # NOTE: Unused in the N-level case
    else:
        moiinputs['y'] = str(y)
        moiinputs['x'] = ",".join(inputlist(x))
        moiinputs['D'] = str(D)
        if D == 'Binomial':
            moiinputs['n'] = "autodenom"
            moiinputs['link'] = str(link)
        if D == 'Poisson':
            moiinputs['link'] = str(link)
            moiinputs['offset'] = str(offset)
            if offset:
                moiinputs['n'] = str(n)

    if templatename == 'NLevelRS':
        for j in range(0, int(NumLevs)):
            moiinputs['x' + str(j+1)] = ",".join(context['x' + str(j+1)])
            moiinputs['priors' + str(j)] = str(context['priorsMOI' + str(j+1)])
            if context['priorsMOI' + str(j+1)] == 'Wishart':
                moiinputs['R' + str(j)] = ",".join(context['RMOI' + str(j+1)])
                moiinputs['v' + str(j)] = str(context['vMOI' + str(j+1)])


#from multiprocessing.pool import ThreadPool
#pool = ThreadPool(processes=len(imputes)+1)

moithread = []
for i in range(0, len(imputes)):
    for j in range(0, len(imputes[i])):
        if templatename == 'RespNLevelMVMixedResponsemvu':
            for lev in range(0, int(NumLevs)):
                if context['MOIanyL'+str(lev+2)+'resp']:
                    moiinputs['L'+str(lev+2)+'Data'] = '_MOIL' + str(lev+2) + 'Data' + str(i) + '_iter' + str(imputes[i][j])

        moithread.append(RunStatJR(template= str(templatename), dataset= 'impute_datafile_chain' + str(i) + '_iter' + str(int(imputefirst)+imputes[i][j]), invars= dict(moiinputs), estoptions={'burnin': str(burnin), 'iterations': str(iterations), 'thinning': '1', 'seed': str(i + 1), 'nchains': '1', 'Engine':'eStat', 'defaultalg':'Yes', 'defaultsv':'Yes', 'makepred':'No', 'outdata':'imp_'+str(i+1)+'_'+str(j)}))

        #moithread.append(pool.apply_async(RunStatJR, kwds = {'template': str(templatename), 'dataset': 'impute_datafile_chain' + str(i) + '_iter' + str(imputes[i][j]), 'invars': dict(moiinputs), 'estoptions':{'burnin': str(burnin), 'iterations': str(iterations), 'thinning': '1', 'seed': str(i + 1), 'nchains': '1', 'Engine':'eStat', 'defaultalg':'Yes', 'defaultsv':'Yes', 'makepred':'No', 'outdata':'imp_'+str(i+1)+'_'+str(j)}}))

if templatename == 'RespNLevelMVMixedResponsemvu':
    for lev in range(0, int(NumLevs)):
        if context['MOIanyL'+str(lev+2)+'resp']:
            moiinputs['L'+str(lev+2)+'Data'] = '_MOIL' + str(lev+2) + 'Data'
#completethread = pool.apply_async(RunStatJR, kwds = {'template': str(templatename), 'dataset': 'CompleteData', 'invars': dict(moiinputs), 'estoptions':{'burnin': str(burnin), 'iterations': str(iterations), 'thinning': '1', 'seed': str(i + 1), 'nchains': '1', 'Engine':'eStat', 'defaultalg':'Yes', 'defaultsv':'Yes', 'makepred':'No', 'outdata':'imp'+str(i+1)}})
completethread = RunStatJR(template =  str(templatename), dataset = 'CompleteData', invars = dict(moiinputs), estoptions = {'burnin': str(burnin), 'iterations': str(iterations), 'thinning': '1', 'seed': str(i + 1), 'nchains': '1', 'Engine':'eStat', 'defaultalg':'Yes', 'defaultsv':'Yes', 'makepred':'No', 'outdata':'imp'+str(i+1)})

#pool.close()

print 'Datasets: ' + str(datasets.datasets.keys())

#pool.join()

moi = []
for i in range(0, len(moithread)):
    #m = moithread[i].get()
    m = moithread[i]
    moi.append(m)

    for k in m.eng.inputs.keys():
        outputs['Model' + str(i + 1) + '_' + k] = moi[i].eng.inputs[k]

    for k in m.eng.outputs.keys():
        outputs['Model' + str(i + 1) + '_' + k] = moi[i].eng.outputs[k]

combinedchains = {}
for p in moi[0].eng.outputs['Chains'].chains.keys():
    chains = moi[0].eng.outputs['Chains'].chains[p]
    combinedchains[p] = numpy.empty((len(moi), chains.size))

meanDIC = 0
meanpD = 0
for i in range(0, len(moi)):
    for p in moi[i].eng.outputs['Chains'].chains.keys():
        chains = moi[i].eng.outputs['Chains'].chains[p]
        combinedchains[p][i, :] = chains
    if 'DIC' in moi[i].eng.outputs['ModelResults'].content['modelstats']:
        meanDIC += moi[i].eng.outputs['ModelResults'].content['modelstats']['DIC']
    if 'pD' in moi[i].eng.outputs['ModelResults'].content['modelstats']:
        meanpD += moi[i].eng.outputs['ModelResults'].content['modelstats']['pD']

meanDIC /= len(moi)
meanpD /= len(moi)

combinedresults = ModelOutput()
for p in combinedchains.keys():
    combinedresults.add(p, 'chains', combinedchains[p])
    outputs['Combined_' + p + '.svg'] = MCMCGraph(p, combinedchains[p], moi[0].eng.directory)
combinedresults.summarisechains()
outputs['CombinedResults'] = combinedresults

#completemoi = completethread.get()
completemoi = completethread
for k in completemoi.eng.inputs.keys():
    outputs['CompleteCasesModel' + '_' + k] = completemoi.eng.inputs[k]

for k in completemoi.eng.outputs.keys():
    outputs['CompleteCasesModel' + '_' + k] = completemoi.eng.outputs[k]

tabout = TabularOutput()
tabout.column_headings = ['parameter', 'mean', 'standard error', 'variable']

if meanDIC != 0:
    tabout.add_row('Mean DIC', [meanDIC, '', ''])
if meanpD != 0:
    tabout.add_row('Mean pD', [meanpD, '', ''])

for p in sorted(moi[0].eng.outputs['ModelResults'].content['paramstats']['mean'].keys()):
    if 'variable' in moi[0].eng.outputs['ModelResults'].content['paramstats'] and p in moi[0].eng.outputs['ModelResults'].content['paramstats']['variable']:
        varname = moi[0].eng.outputs['ModelResults'].content['paramstats']['variable'][p]
    else:
        varname = ''

    thetahat = 0
    for i in range(0, len(moi)):
        thetahat += moi[i].eng.outputs['ModelResults'].content['paramstats']['mean'][p]
    thetahat /= len(moi)

    vbar = 0
    for i in range(0, len(moi)):
        vbar += numpy.power(moi[i].eng.outputs['ModelResults'].content['paramstats']['sd'][p], 2)
    vbar /= len(moi)

    B = 0
    for i in range(0, len(moi)):
        B += numpy.power((moi[i].eng.outputs['ModelResults'].content['paramstats']['mean'][p] - thetahat), 2) / (len(moi) - 1)

    vartheta = vbar + (1 + (1/len(moi)))*B

    if not p.startswith('tau') and not p=='deviance':
        tabout.add_row(p, [thetahat, numpy.sqrt(vartheta), varname])

outputs['ResultsTable:Imputation'] = tabout

taboutcomp = TabularOutput()
taboutcomp.column_headings = ['parameter', 'mean', 'standard error', 'variable']
if 'DIC' in completemoi.eng.outputs['ModelResults'].content['modelstats']:
    taboutcomp.add_row('DIC', [completemoi.eng.outputs['ModelResults'].content['modelstats']['DIC'], '', ''])
if 'pD' in completemoi.eng.outputs['ModelResults'].content['modelstats']:
    taboutcomp.add_row('pD', [completemoi.eng.outputs['ModelResults'].content['modelstats']['pD'], '', ''])
for p in sorted(completemoi.eng.outputs['ModelResults'].content['paramstats']['mean'].keys()):
    if 'variable' in completemoi.eng.outputs['ModelResults'].content['paramstats'] and p in completemoi.eng.outputs['ModelResults'].content['paramstats']['variable']:
        varname = completemoi.eng.outputs['ModelResults'].content['paramstats']['variable'][p]
    else:
        varname = ''

    meancomplete = completemoi.eng.outputs['ModelResults'].content['paramstats']['mean'][p]
    varcomplete = numpy.power(completemoi.eng.outputs['ModelResults'].content['paramstats']['sd'][p], 2)
    if not p.startswith('tau') and not p=='deviance':
        taboutcomp.add_row(p, [meancomplete, numpy.sqrt(varcomplete), varname])

outputs['ResultsTable:CompleteCases'] = taboutcomp
'''
