# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateRespNLevelMVMixedResponsemvu(Template):
    'Fits models with multivariate mixed response models (Normal, Binary, ordered, or unordered), allowing for missing data and a choice of uniform or Wishart priors and with the responses present at N levels. Creates custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'Normal', 'Binary', 'Ordered multinomial', 'Unordered multinomial', 'Multivariate response']
    engines = ['CustomC']

    inputs = '''
y1 = DataMatrix('Responses: ')
Nhigherlevels = Integer ('Number of higher levels')
endlev = Nhigherlevels + 2
for lev in range(2, endlev):
    context['L'+str(lev)+'ID'] = IDVector('Level ' + str(lev)+' ID: ')

for i in range(0, len(y1)):
    context['bin1_'+str(i+1)] = Text('Response Type for response ' + y1[i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
    if(context['bin1_'+str(i+1)] == 'Ordered'):
       context['ncats'+y1[i]] = Integer('Number of categories for ' + y1[i] + ': ')
    if(context['bin1_'+str(i+1)] == 'Unordered'):
       context['ncats'+y1[i]] = Integer('Number of categories for ' + y1[i] + ': ')
       for j in range(0, context['ncats'+y1[i]]-1):
          context['x1_'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
    else:
       context['x1_'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y1[i] + ': ', allow_cat = True)

for lev in range(2, endlev):
    context['anyL'+str(lev)+'resp'] = Boolean('Are there any Level ' + str(lev) + ' responses:')
    if context['anyL'+str(lev)+'resp']:
        context['L'+str(lev)+'Data'] = DataFileChooser('Level ' + str(lev)+' data set') 
        context['y'+str(lev)] = DataMatrix('Responses at level'+str(lev)+': ', data_src= context['L'+str(lev)+'Data'])
        for i in range(0, len(context['y'+str(lev)])):
            context['bin'+str(lev)+'_'+str(i+1)] = Text('Response Type for response ' + context['y' + str(lev)][i] + ': ', ['Normal', 'Binary', 'Ordered', 'Unordered'])
            if context['bin'+str(lev)+'_'+str(i+1)] == 'Ordered':
                context['ncats'+context['y'+str(lev)][i]] = Integer('Number of categories for ' + context['y'+str(lev)][i] + ': ')
            if context['bin'+str(lev)+'_'+str(i+1)] == 'Unordered':
                context['ncats'+context['y'+str(lev)][i]] = Integer('Number of categories for ' + context['y'+str(lev)][i] + ': ')
                for j in range(0, context['ncats'+context['y'+str(lev)][i]]-1):
                    context['x'+str(lev)+'_'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', data_src= context['L'+str(lev)+'Data'], allow_cat = True)
            else:
                context['x' +str(lev)+'_'+str(i+1)] = DataMatrix('Explanatory variables for response ' + context['y'+str(lev)][i] + ': ', data_src= context['L'+str(lev)+'Data'], allow_cat = True)

condmarg = Boolean('Use conditional marginal algorithm: ')
storeresid = Boolean('Store residuals?')

lenbeta1 = 0
n1 = 0
uind = 0
for i in range(0, len(y1)):
    if(context['bin1_'+str(i+1)] == 'Normal'):
       context['miss'+y1[i]] = ParamVector(monitor=False)
    if(context['bin1_'+str(i+1)] == 'Binary'):
       context['misb'+y1[i]] = ParamVector(monitor=False)
    if(context['bin1_'+str(i+1)] == 'Ordered'):
       context['miso'+y1[i]] = ParamVector(monitor=False)
       context['alpha'+y1[i]] = ParamVector()
       context['alpha'+y1[i]].ncols = -(context['ncats'+y1[i]] -1)
       context['alpha'+y1[i]].set_value( map(str, range(0, (context['ncats'+y1[i]] -1))) )
    if(context['bin1_'+str(i+1)] == 'Unordered'):
       for j in range(0, context['ncats'+y1[i]]-1):
          lenbeta1 += len(context['x1_'+str(i+1)+'_'+str(j)])
          context['misu'+str(context['ncats'+y1[i]])+'_'+y1[i]+'_'+ str(j)] = ParamVector(monitor=False)
          for lev in range(2, endlev):
              if storeresid:
                  context['u' + str(lev) + '_' + str(uind)] = ParamVector(parents=[context['L'+str(lev)+'ID']], as_scalar=False)
              else:
                  context['u' + str(lev) + '_' + str(uind)] = ParamVector(parents=[context['L'+str(lev)+'ID']], as_scalar=False, monitor=False)
          uind += 1
       context['y1_'+ str(i+1)] = DataMatrix(value=[])
       context['origv1_'+str(i+1)] = List(value = [])
       n1 += context['ncats'+y1[i]] - 1
    else:
       lenbeta1 += len(context['x1_'+str(i+1)])
       n1 += 1
       for lev in range(2, endlev):
           if storeresid:
               context['u' + str(lev) + '_' + str(uind)] = ParamVector(parents=[context['L'+str(lev)+'ID']], as_scalar=False)
           else:
               context['u' + str(lev) + '_' + str(uind)] = ParamVector(parents=[context['L'+str(lev)+'ID']], as_scalar=False, monitor=False)
       uind += 1

omega_e1 = ParamMatrix()
omega_e1.size = n1
omega_e1.select('Starting values for Level 1 variance matrix (lower triangle): ')
d_e1 = ParamMatrix(modelled = False)
d_e1.size = n1
beta1 = ParamVector()
beta1.ncols = lenbeta1

for lev in range(2, endlev):
    lenbeta2 = 0 
# only temporary so leave as 2
    n2 = 0
# only temporary so leave as 2
    if context['anyL'+str(lev)+'resp']:
        for i in range(0, len(context['y'+str(lev)])):
            if(context['bin'+str(lev)+'_'+str(i+1)] == 'Normal'):
               context['miss'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]] = ParamVector(monitor=False)
            if(context['bin'+str(lev)+'_'+str(i+1)] == 'Binary'):
               context['misb'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]] = ParamVector(monitor=False)
            if(context['bin'+str(lev)+'_'+str(i+1)] == 'Ordered'):
               context['miso'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]] = ParamVector(monitor=False)
               context['alpha'+context['y'+str(lev)][i]] = ParamVector()
               context['alpha'+context['y'+str(lev)][i]].ncols = -(context['ncats'+context['y'+str(lev)][i]] -1)
               context['alpha'+context['y'+str(lev)][i]].set_value( map(str, range(0, (context['ncats'+context['y'+str(lev)][i]] -1))) )
            if(context['bin'+str(lev)+'_'+str(i+1)] == 'Unordered'):
               for j in range(0, context['ncats'+context['y'+str(lev)][i]]-1):
                   lenbeta2 += len(context['x'+str(lev)+'_'+str(i+1)+'_'+str(j)])
                   context['misu'+str(context['L'+str(lev)+'Data'])+'_'+str(context['ncats'+context['y'+str(lev)][i]])+'_'+context['y'+str(lev)][i]+'_'+ str(j)] = ParamVector(monitor=False)
               context['y'+str(lev)+'_'+ str(i+1)] = DataMatrix(value=[], data_src= context['L'+str(lev)+'Data'] )
               context['origv'+str(lev)+'_'+str(i+1)] = Text(value = [])
               n2 += context['ncats'+context['y'+str(lev)][i]]- 1
            else:
               lenbeta2 += len(context['x' + str(lev) +'_'+str(i+1)])
               n2 += 1

        context['beta'+str(lev)] = ParamVector()
        context['beta'+str(lev)].ncols = lenbeta2

    context['omega_u' + str(lev)] = ParamMatrix(modelled = False)
    context['omega_u' + str(lev)].size = n2 + n1
    context['d_u' + str(lev)] = ParamMatrix()
    context['d_u' + str(lev)].size = n2 + n1

deviance = ParamScalar(modelled = False)
'''

    customccode = '''
        <%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += name + '[' + index + '] * ' + var + '[' + str(count) + ']'
            count += 1
        return out

    def PyToC(var):
        if isinstance(var, list):
            return "{" + ",".join(map(PyToC, var)) + "}"
        else:
            return str(var)

    def mmult3(dset, names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += str(dset) + '_' + name + '[' + index + '] * ' + var + '[' + str(count) + ']'
            count += 1
        return out
%>

<%
lvars = {}
nvar1 = len(y1)
n1 = 0
for i in range(0, nvar1):
    if (context['bin1_'+str(i+1)] == 'Unordered'):
        n1 += context['ncats'+y1[i]] - 1
    else:
        n1 += 1

endlev = Nhigherlevels + 2
for lev in range(2, endlev):
    lvars['n' + str(lev)] = 0
    lvars['nvar' + str(lev)] = 0
    if context['anyL'+str(lev)+'resp']:
        lvars['nvar' + str(lev)] = len(context['y' + str(lev)])
        for i in range(0, lvars['nvar' + str(lev)]):
            if (context['bin' + str(lev) +  '_' + str(i+1)] == 'Unordered'):
                lvars['n' + str(lev)] += context['ncats' + context['y' + str(lev)][i]] - 1
            else:
                lvars['n' + str(lev)] += 1

    lvars['nr' + str(lev)] = lvars['n' + str(lev)] + n1 # nrn is total number of random effects at level n

lenbeta1 = 0
for i in range(0, nvar1):
    if (context['bin1_'+str(i+1)] == 'Unordered'):
        for j in range(0,context['ncats'+y1[i]]-1):
            lenbeta1 += len(context['x1_' + str(i + 1) + '_'+ str(j)])
    else:
        lenbeta1 += len(context['x1_' + str(i + 1)])

for lev in range(2, endlev):
    lvars['lenbeta' + str(lev)] = 0
    if context['anyL'+str(lev)+'resp']:
        for i in range(0, lvars['nvar' + str(lev)]):
            if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
                for j in range(0, context['ncats' + context['y' + str(lev)][i]]-1):
                    lvars['lenbeta' + str(lev)] += len(context['x' + str(lev) + '_' + str(i + 1) + '_'+ str(j)])
            else:
                lvars['lenbeta' + str(lev)] += len(context['x' + str(lev) + '_' + str(i + 1)])

allnormal1 = True
for i in range(0, nvar1):
    if (context['bin1_' + str(i+1)] != 'Normal'):
        allnormal1 = False

for lev in range(2, endlev):
    lvars['allnormal' + str(lev)] = True
    for i in range(0, lvars['nvar' + str(lev)]):
        if (context['bin' + str(lev) + '_' + str(i+1)] != 'Normal'):
            lvars['allnormal' + str(lev)] = False

pyyind1 = []
pyymap1 = []
pypos1 = 0
for i in range(0, len(y1)):
    if (context['bin1_'+str(i+1)] == 'Unordered'):
        pyymap1.append(pypos1)
        pypos1 += context['ncats'+y1[i]]-1
        for j in range(0,context['ncats'+y1[i]]-1):
            pyyind1.append(i)
    else:
        pyymap1.append(pypos1)
        pypos1 += 1
        pyyind1.append(i)

for lev in range(2, endlev):
    lvars['pyyind' + str(lev)] = []
    lvars['pyymap' + str(lev)] = []
    lvars['pypos' + str(lev)] = 0
    if context['anyL'+str(lev)+'resp']:
        for i in range(0, len(context['y' + str(lev)])):
            if (context['bin' + str(lev) + '_'+str(i+1)] == 'Unordered'):
                lvars['pyymap' + str(lev)].append(lvars['pypos' + str(lev)])
                for j in range(0,context['ncats'+context['y' + str(lev)][i]]-1):
                    lvars['pypos' + str(lev)] += 1
                    lvars['pyyind' + str(lev)].append(i)
            else:
                lvars['pyymap' + str(lev)].append(lvars['pypos' + str(lev)])
                lvars['pypos' + str(lev)] += 1
                lvars['pyyind' + str(lev)].append(i)
%>

                std::vector<int> betaind1(${lenbeta1});
<% betacount = 0 %>
<% row = 0 %>\\
% for i in range(0, len(y1)):
% if (context['bin1_'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y1[i]]-1):
% for k in range(0, len(context['x1_'+str(i+1)+'_'+str(j)])):
                betaind1[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endfor
% else:
% for j in range(0, len(context['x1_' +str(i+1)])):
                betaind1[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endif
% endfor

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
                std::vector<int> betaind${lev}(${lvars['lenbeta' + str(lev)]});
<% betacount = 0 %>
<% row = 0 %>\\
% for i in range(0, len(context['y' + str(lev)])):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+context['y' + str(lev)][i]]-1):
% for k in range(0, len(context['x' + str(lev) + '_' + str(i+1) + '_' + str(j)])):
                betaind${lev}[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endfor
% else:
% for j in range(0, len(context['x' + str(lev) + '_' + str(i+1)])):
                betaind${lev}[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endif
% endfor
% endif
% endfor

                std::vector<int> betamap1(${n1});
                int pos1 = 0;
<% ycount = 0 %>
% for i in range(0, len(y1)):
% if (context['bin1_'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y1[i]]-1):
                betamap1[${ycount}] = pos1;
                pos1 += mat_x1_${i + 1}_${j}.cols();
<% ycount += 1 %>
% endfor
% else:
                betamap1[${ycount}] = pos1;
                pos1 += mat_x1_${i + 1}.cols();
<% ycount += 1 %>
% endif
% endfor

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
                std::vector<int> betamap${lev}(${lvars['n' + str(lev)]});
                int pos${lev} = 0;
<% ycount = 0 %>
% for i in range(0, len(context['y' + str(lev)])):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+context['y' + str(lev)][i]]-1):
                betamap${lev}[${ycount}] = pos${lev};
                pos${lev} += mat_x${lev}_${i + 1}_${j}.cols();
<% ycount += 1 %>
% endfor
% else:
                betamap${lev}[${ycount}] = pos${lev};
                pos${lev} += mat_x${lev}_${i + 1}.cols();
<% ycount += 1 %>
% endif
% endfor
% endif
% endfor

                std::vector<double*> tmp_x1;
% for i in range(0, len(y1)):
% if (context['bin1_'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y1[i]]-1):
% for k in range(0, len(context['x1_'+str(i+1)+'_'+str(j)])):
                tmp_x1.push_back(${context['x1_' + str(i+1)+'_'+str(j)][k]});
% endfor
% endfor
% else:
% for j in range(0, len(context['x1_' + str(i+1)])):
                tmp_x1.push_back(${context['x1_' + str(i+1)][j]});
% endfor
% endif
% endfor
                RectMatrix mat_x1(tmp_x1, length(${y1[0]}));

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
                std::vector<double*> tmp_x${lev};
% for i in range(0, len(context['y' + str(lev)])):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats' + context['y' + str(lev)][i]]-1):
% for k in range(0, len(context['x' + str(lev) + '_' + str(i+1) + '_' + str(j)])):
                tmp_x${lev}.push_back(${context['L' + str(lev) + 'Data']}_${context['x' + str(lev) + '_' + str(i+1) + '_' + str(j)][k]});
% endfor
% endfor
% else:
% for j in range(0, len(context['x' + str(lev) + '_' + str(i+1)])):
                tmp_x${lev}.push_back(${context['L' + str(lev) + 'Data']}_${context['x' + str(lev) + '_' + str(i+1)][j]});
% endfor
% endif
% endfor
                RectMatrix mat_x${lev}(tmp_x${lev}, length(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][0]}));
% endif
% endfor

                std::vector<double*> tmp_ymiss1;

% for i in range(0, len(y1)):
% if (context['bin1_'+str(i+1)] == 'Normal'):
                tmp_ymiss1.push_back(miss${y1[i]});
% endif
% if (context['bin1_'+str(i+1)] == 'Binary'):
                tmp_ymiss1.push_back(misb${y1[i]});
% endif
% if (context['bin1_'+str(i+1)] == 'Ordered'):
                tmp_ymiss1.push_back(miso${y1[i]});
% endif
% if (context['bin1_'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y1[i]]-1):
                tmp_ymiss1.push_back(misu${context['ncats'+y1[i]]}_${y1[i]}_${j});
% endfor
% endif
% endfor
                RectMatrix mat_ymiss1(tmp_ymiss1, length(${y1[0]}));

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
                std::vector<double*> tmp_ymiss${lev};
% for i in range(0, len(context['y' + str(lev)])):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Normal'):
                tmp_ymiss${lev}.push_back(miss${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]});
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Binary'):
                tmp_ymiss${lev}.push_back(misb${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]});
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Ordered'):
                tmp_ymiss${lev}.push_back(miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]});
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+context['y' + str(lev)][i]]-1):
                tmp_ymiss${lev}.push_back(misu${context['L' + str(lev) + 'Data']}_${context['ncats'+context['y' + str(lev)][i]]}_${context['y' + str(lev)][i]}_${j});
% endfor
% endif
% endfor
                RectMatrix mat_ymiss${lev}(tmp_ymiss${lev}, length(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][0]}));
% endif
% endfor
% for lev in range(2, endlev):
                std::vector<double*> tmp_ul${lev};
% for i in range(0, n1):
                tmp_ul${lev}.push_back(u${lev}_${i});
% endfor
                RectMatrix mat_ul${lev}(tmp_ul${lev}, length(u${lev}_0));
% endfor

% for var in range(0, nvar1):
% if (context['bin1_'+str(var+1)] == 'Ordered'):
    const int ncats${y1[var]} = ${context['ncats'+y1[var]]};
% endif  
% if (context['bin1_'+str(var+1)] == 'Unordered'):
    const int ncats${y1[var]} = ${context['ncats'+y1[var]]};
% endif  
% endfor

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
% for var in range(0, lvars['nvar' + str(lev)]):
% if (context['bin' + str(lev) + '_'+str(var+1)] == 'Ordered'):
    const int ncats${context['y' + str(lev)][var]} = ${context['ncats'+context['y' + str(lev)][var]]};
% endif  
% if (context['bin' + str(lev) + '_'+str(var+1)] == 'Unordered'):
    const int ncats${context['y' + str(lev)][var]} = ${context['ncats'+context['y' + str(lev)][var]]};
% endif  
% endfor
% endif
% endfor



    {
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for i in range(0, nvar1):
% if (context['bin1_'+str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+y1[i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+y1[i]]))};

            for(int j = 0; j < length(${y1[i]}); j++) {
                if(${y1[i]}[j] <= -9.999e29) {
                } else {                
                    if(miso${y1[i]}[j] < low[int(${y1[i]}[j])])
                        low[int(${y1[i]}[j])] = miso${y1[i]}[j];
                    if(miso${y1[i]}[j] > upp[int(${y1[i]}[j])])
                        upp[int(${y1[i]}[j])] = miso${y1[i]}[j];
                }
            } 
            for(int j = 1; j<ncats${y1[i]}-1; j++) {
                double ul,ll;            
                if(upp[j]>double(alpha${y1[i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${y1[i]}[j-1]);
                if(j == (ncats${y1[i]}-2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${y1[i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${y1[i]}[j+1]);
                alpha${y1[i]}[j]= ll+(ul-ll)*random01();
            }
        }
% endif
% endfor
        // Bill's attempt of step for d_e

        // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb1(${n1});

        for(int i = 0; i < length(${y1[0]}); i++) {
<% lenbeta1 = 0 %>\\
<% row = 0 %>\\
<% uind = 0 %>\\
% for i in range(0, nvar1):
% if (context['bin1_'+str(i+1)] == 'Normal'):
            double e${row} = miss${y1[i]}[i] - (${mmult2(context['x1_' + str(i+1)], 'beta1', 'i', lenbeta1)})\\
% for lev in range(2, endlev):
 - u${lev}_${uind}[${context['L'+str(lev)+'ID']}[i]] \\
% endfor
;
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin1_'+str(i+1)] == 'Binary'):
            double e${row} = misb${y1[i]}[i] - (${mmult2(context['x1_' + str(i+1)], 'beta1', 'i', lenbeta1)})\\
% for lev in range(2, endlev):
 - u${lev}_${uind}[${context['L'+str(lev)+'ID']}[i]] \\
% endfor
;
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin1_'+str(i+1)] == 'Ordered'):
            double e${row} = miso${y1[i]}[i] - (${mmult2(context['x1_' + str(i+1)], 'beta1', 'i', lenbeta1)})\\
% for lev in range(2, endlev):
 - u${lev}_${uind}[${context['L'+str(lev)+'ID']}[i]] \\
% endfor
;
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin1_'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y1[i]]-1):
            double e${row} = misu${context['ncats'+y1[i]]}_${y1[i]}_${j}[i] - (${mmult2(context['x1_' + str(i+1)+'_'+str(j)], 'beta1', 'i', lenbeta1)})\\
% for lev in range(2, endlev):
 - u${lev}_${uind}[${context['L'+str(lev)+'ID']}[i]] \\
% endfor
;
<% uind += 1 %>\\
<% row += 1 %>\\
<% lenbeta1 += len(context['x1_' + str(i + 1) + '_'+ str(j)]) %>\\
% endfor
% else :
<% lenbeta1 += len(context['x1_' + str(i + 1)]) %>\\
% endif
% endfor
% for i in range(0, n1):
% for j in range(i, n1):
            sb1(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }

        bool fail = false;
% if allnormal1:
        int vwe = mat_ymiss1.rows() - ${n1 + 1};
        matrix_sym_invinplace(sb1);
        mat_d_e1 = dwishart(vwe, sb1, fail);
        mat_omega_e1 = matrix_sym_inverse(mat_d_e1, fail);
% else:
        if (runstate == 0)
            mat_d_e1 = matrix_sym_inverse(mat_omega_e1, fail);

        // Now loop over elements of va (omega_e)
        double llcurr;
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, n1):
<% var2 = var1 %>\\
<% withinvar2 = withinvar1 %>\\
% for j in range(i, n1):
        {
% if var1 == var2 and (context['bin1_'+str(var1+1)] != 'Normal'):
            omega_e1_accept[${count}] = 44;
% else:
            llcurr = length(${y1[0]})*log(matrix_determinant(mat_d_e1))/2;
            for(int i1 = 0; i1 < ${n1}; i1++)
                for(int j1 = 0; j1 < ${n1}; j1++)
                    llcurr -= sb1(i1, j1) * mat_d_e1(i1, j1)/2;
            double old = mat_omega_e1(${i}, ${j});
            mat_omega_e1(${i}, ${j}) = dnorm(mat_omega_e1(${i}, ${j}), 1.0/sqrt(omega_e1_sd[${count}]));
            // mat_d_e1 = matrix_sym_inverse(mat_omega_e1, fail);
            SymMatrix temp_d_e1 = matrix_cholesky(mat_omega_e1, fail);
            if (fail == true) {
                mat_omega_e1(${i}, ${j}) = old;
                // mat_d_e1 = matrix_sym_inverse(mat_omega_e1, fail);
            } else {
                temp_d_e1 = matrix_cholinverse(temp_d_e1);
                double llnew = length(${y1[0]})*log(matrix_determinant(temp_d_e1))/2;
                for(int i1 = 0; i1 < ${n1}; i1++)
                    for(int j1 = 0; j1 < ${n1}; j1++)
                        llnew -= sb1(i1, j1) * temp_d_e1(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                    if(random01() > exp(llnew-llcurr))
                        change = 0;
                }
                if(change == 0) {
                    mat_omega_e1(${i}, ${j}) = old;
                    // mat_d_e1 = matrix_sym_inverse(mat_omega_e1, fail);
                } else {
                    omega_e1_accept[${count}]++;
                    mat_d_e1 = temp_d_e1;
                }
            }
% endif
<% count += 1 %>\\
        }
% if (context['bin1_'+str(var2+1)] != 'Unordered'):
<% var2+= 1 %>\\
% elif withinvar2 != (context['ncats'+y1[var2]]-2):
<% withinvar2 += 1 %>\\
%else:
<% var2+= 1 %>\\
<% withinvar2 = 0 %>\\
% endif
% endfor
% if (context['bin1_'+str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+y1[var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif
% endfor

% endif
     
                // Step for beta

                SymMatrix db1(${lenbeta1});
                for (int i = 0; i < length(${y1[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta1}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db1(j1, j2) += mat_x1(i, j1) * mat_d_e1(betaind1[j1], betaind1[j2]) * mat_x1(i, j2);
                        }
                    }
                }

                RectMatrix xtdy1(${lenbeta1}, 1);
                for (int i = 0; i < length(${y1[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta1}; j1++) {
                        for (int j2 = 0; j2 < ${n1}; j2++) {
                            xtdy1(j1, 0) += mat_x1(i, j1) * mat_d_e1(betaind1[j1], j2) * (mat_ymiss1(i, j2)\\
% for lev in range(2, endlev):
 - mat_ul${lev}(${context['L'+str(lev)+'ID']}[i], betaind1[j2]) \\
% endfor                            
);
                        }
                    }
                }

                fail = false;
                // int err = matrix_sym_invinplace(db1);
                // RectMatrix betahat1 = db1 * xtdy1;
                // mat_beta1 = dmultnormal(betahat1, db1, fail);

                RectMatrix betahat1 = matrix_solve(db1, xtdy1);
                mat_beta1 = dmultnormalinv(betahat1, db1, fail);


		// Step for u's (Multivariate!!!!)
% for lev1 in range(2, endlev):
                for (int j = 0; j < mat_ul${lev1}.rows(); j++) {
                    SymMatrix du1(${n1});
                    for(int j1 = 0; j1 < ${n1}; j1++) {
                        for(int j2 = 0;j2 < ${n1}; j2++) {
                            du1(j1, j2) = mat_d_e1(j1, j2)*idmap_${context['L'+str(lev1)+'ID']}[j].size() + mat_d_u${lev1}(${lvars['n' + str(lev1)]} + j1, ${lvars['n' + str(lev1)]} + j2);
                        }
                    }
                
                    std::vector<double> sumx(mat_x1.cols());
                    std::vector<double> sumy(mat_ymiss1.cols());
% for lev2 in range(2, endlev):
% if lev2 != lev1:
                    std::vector<double> sumu${lev2}(mat_ul${lev2}.cols());
% endif
% endfor
                    for (int id_k = 0; id_k < idmap_${context['L'+str(lev1)+'ID']}[j].size(); id_k++) {
                        int k = idmap_${context['L'+str(lev1)+'ID']}[j][id_k];
                        for (int m = 0; m < mat_x1.cols(); m++) {
                            sumx[m] += mat_x1(k, m);
                        }
                        for (int m = 0; m < mat_ymiss1.cols(); m++) {
                            sumy[m] += mat_ymiss1(k, m);
                        }
% for lev2 in range(2, endlev):
% if lev2 != lev1:
                        for (int m = 0; m < mat_ul${lev2}.cols(); m++) {
                            sumu${lev2}[m] += mat_ul${lev2}(${context['L'+str(lev2)+'ID']}[k], m);
                        }
% endif
% endfor
                    }

                    std::vector<double> sumymxb(mat_ymiss1.cols());
                    for (int m = 0; m < mat_ymiss1.cols(); m++) {
                        sumymxb[m] = sumy[m];
                    }
                    for (int k = 0; k < mat_x1.cols(); k++) {
                        sumymxb[betaind1[k]] -= mat_beta1(k, 0) * sumx[k]; 
                    }

% for lev2 in range(2, endlev):
% if lev2 != lev1:
                    for (int m = 0; m < mat_ul${lev2}.cols(); m++) {
                        sumymxb[m] -= sumu${lev2}[m];
                    }
% endif
% endfor

                    RectMatrix xtdu1(${n1}, 1);
                    for(int j1 = 0; j1 < ${n1}; j1++) {
                        for(int j2 = 0;j2 < ${n1}; j2++) {
                            xtdu1(j1, 0) += mat_d_e1(j1, j2) * sumymxb[j2];
                        }
                    }

% if context['anyL'+str(lev1)+'resp']:
                    std::vector<double> sumymxb2(${lvars['n' + str(lev1)]});
                    for (int m = 0; m < mat_ymiss${lev1}.cols(); m++) {
                        sumymxb2[m] = mat_ymiss${lev1}(j, m);
                    }
                    for (int k = 0; k < mat_x${lev1}.cols(); k++) {
                        sumymxb2[betaind${lev1}[k]] -= mat_beta${lev1}(k, 0) * mat_x${lev1}(j, k); 
                    }

                    for(int j1 = 0;j1 < ${n1}; j1++) {
                        for(int j2 = 0; j2 < ${lvars['n' + str(lev1)]}; j2++) {
                            xtdu1(j1, 0) += mat_d_u${lev1}(${lvars['n' + str(lev1)]} + j1, j2) * sumymxb2[j2];     
                        }
                    }
% endif

                    fail = false;
                    RectMatrix uhat = matrix_solve(du1, xtdu1);
                    RectMatrix mat_u_temp = dmultnormalinv(uhat, du1, fail);

                    for (int j1 = 0; j1 < ${n1}; j1++) {
                        mat_ul${lev1}(j, j1) = mat_u_temp(j1, 0);
                    }
		}
% endfor


% for i in range(0, n1):

<%
mat_x_i = 'mat_x1_' + str(pyyind1[i] + 1)
if (context['bin1_'+str(pyyind1[i]+1)] == 'Unordered'):
    mat_x_i += '_' + str(i - pyymap1[pyyind1[i]])
%>

                for (int j = 0; j < mat_ymiss1.rows(); j++) {
                    //if (mat_y1(j, ${pyyind1[i]}) <= -9.999e29) {
                    {
                        double tmp = 0;
% for lev in range(0, Nhigherlevels):
                        tmp += mat_u${lev+2}_${i}(${context['L' + str(lev+2) + 'ID']}[j], 0);
% endfor
                        for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                            tmp += mat_beta1(betamap1[${i}] + k, 0) * ${mat_x_i}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss1.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e1(${i}, k) * mat_ymiss1(j, k) * (1.0 / mat_d_e1(${i}, ${i}));
                            }
                        }
% for m in range(0, n1):
% if i != m:
<%
mat_x_m = 'mat_x1_' + str(pyyind1[m] + 1)
if (context['bin1_'+str(pyyind1[m]+1)] == 'Unordered'):
    mat_x_m += '_' + str(m - pyymap1[pyyind1[m]])
%>
% for lev in range(0, Nhigherlevels):
                        tmp += mat_d_e1(${i}, ${m}) * mat_u${lev+2}_${m}(${context['L' + str(lev+2) + 'ID']}[j], 0) * (1.0 / mat_d_e1(${i}, ${i}));
% endfor
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp += mat_d_e1(${i}, ${m}) * (1.0 / mat_d_e1(${i}, ${i})) * mat_beta1(betamap1[${m}] + k, 0) * ${mat_x_m}(j, k);
                        }
% endif
% endfor


% if (context['bin1_'+str(pyyind1[i]+1)] == 'Normal'):
                    if (mat_y1(j, ${pyyind1[i]}) <= -9.999e29) {
                        mat_ymiss1(j, ${i}) = dnorm(tmp, mat_d_e1(${i}, ${i}));
                    }
% endif
% if (context['bin1_'+str(pyyind1[i]+1)] == 'Binary'):
                    if (mat_y1(j, ${pyyind1[i]}) <= -9.999e29) 
                        mat_ymiss1(j, ${i}) = dnorm(tmp, mat_d_e1(${i}, ${i}));
                    else if(mat_y1(j, ${pyyind1[i]}) <= 0) 
                        mat_ymiss1(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e1(${i}, ${i})), 2, 0, 0);
                    else          
                        mat_ymiss1(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e1(${i}, ${i})), 1, 0, 0);
% endif
% if (context['bin1_'+str(pyyind1[i]+1)] == 'Ordered'):
                    if (mat_y1(j, ${pyyind1[i]}) <= -9.999e29) 
                        mat_ymiss1(j, ${i}) = dnorm(tmp, mat_d_e1(${i}, ${i}));
                    else if(mat_y1(j, ${pyyind1[i]}) == 0) 
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 2, 0, 0);
                    else if(mat_y1(j, ${pyyind1[i]}) == ${context['ncats'+y1[pyyind1[i]]] - 1})
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 1, alpha${y1[pyyind1[i]]}[ncats${y1[pyyind1[i]]}-2], 0);
                    else          
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 3, alpha${y1[pyyind1[i]]}[int(mat_y1(j, ${pyyind1[i]}))-1], alpha${y1[pyyind1[i]]}[int(mat_y1(j, ${pyyind1[i]}))]);

% endif

% if (context['bin1_'+str(pyyind1[i]+1)] == 'Unordered'):
                    if (mat_y1(j, ${pyyind1[i]}) <= -9.999e29) 
                        mat_ymiss1(j, ${i}) = dnorm(tmp, mat_d_e1(${i}, ${i}));
                    else if (mat_y1(j, ${pyyind1[i]}) == ${context['ncats'+y1[pyyind1[i]]] - 1}) 
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 2, 0, 0);
                    else if (mat_y1(j, ${pyyind1[i]}) == ${i - pyymap1[pyyind1[i]]}) {
                        double limit = 0;
                        for (int k = 0; k < ${context['ncats'+y1[pyyind1[i]]] - 1}; k++) {
                            if ((k != ${i - pyymap1[pyyind1[i]]}) && (mat_ymiss1(j, ${pyymap1[pyyind1[i]]} + k) > limit))
                                limit = mat_ymiss1(j, ${pyymap1[pyyind1[i]]} + k); 
                        }
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 1, limit, 0);
                    } else {  
                        double limit = mat_ymiss1(j, ${pyymap1[pyyind1[i]]});
                        for (int k = 0; k < ${context['ncats'+y1[pyyind1[i]]] - 1}; k++) {
                            if (mat_ymiss1(j, ${pyymap1[pyyind1[i]]} + k) > limit)
                                limit = mat_ymiss1(j, ${pyymap1[pyyind1[i]]} + k); 
                        }
                        mat_ymiss1(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e1(${i}, ${i})), 2, 0, limit);
                    }

% endif

                    }
                }
% endfor






















               // Level 2 code

    {
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for lev in range(2, endlev):
% for i in range(0, lvars['nvar' + str(lev)]):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+context['y' + str(lev)][i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+context['y' + str(lev)][i]]))};

            for(int j = 0; j < length(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}); j++) {
                if (${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j] <= -9.999e29) {
                } else {
                    if(miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j] < low[int(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j])])
                        low[int(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j])] = miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j];
                    if(miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j] > upp[int(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j])])
                        upp[int(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j])] = miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[j];
		}
            } 
            for(int j = 1; j < ncats${context['y' + str(lev)][i]}-1; j++) {
                double ul, ll;            
                if(upp[j]>double(alpha${context['y' + str(lev)][i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${context['y' + str(lev)][i]}[j-1]);
                if(j == (ncats${context['y' + str(lev)][i]}-2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${context['y' + str(lev)][i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${context['y' + str(lev)][i]}[j+1]);
                alpha${context['y' + str(lev)][i]}[j]= ll + (ul-ll) * random01();
            }
        }
% endif
% endfor
% endfor
        // Bill's attempt of step for d_e
             
        bool fail = false;
% for lev in range(2, endlev):
        
        // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb${lev}(${lvars['nr' + str(lev)]});
        for(int i = 0; i < mat_ul${lev}.rows(); i++) {
<% lvars['lenbeta' + str(lev)] = 0 %>\\
<% row = 0 %>\\
% if context['anyL'+str(lev)+'resp']:
% for i in range(0, lvars['nvar' + str(lev)]):
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Normal'):
            double e${row} = miss${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[i] - (${mmult3(context['L' + str(lev) + 'Data'], context['x' + str(lev) + '_' + str(i+1)], 'beta' + str(lev), 'i', lvars['lenbeta' + str(lev)])});
<% row += 1 %>\\
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Binary'):
            double e${row} = misb${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[i] - (${mmult3(context['L' + str(lev) + 'Data'], context['x' + str(lev) + '_' + str(i+1)], 'beta' + str(lev), 'i', lvars['lenbeta' + str(lev)])});
<% row += 1 %>\\
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Ordered'):
            double e${row} = miso${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][i]}[i] - (${mmult3(context['L' + str(lev) + 'Data'], context['x' + str(lev) + '_' + str(i+1)], 'beta' + str(lev), 'i', lvars['lenbeta' + str(lev)])});
<% row += 1 %>\\
% endif
% if (context['bin' + str(lev) + '_' + str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+context['y' + str(lev)][i]]-1):
            double e${row} = misu${context['L' + str(lev) + 'Data']}_${context['ncats'+context['y' + str(lev)][i]]}_${context['y' + str(lev)][i]}_${j}[i] - (${mmult3(context['L' + str(lev) + 'Data'], context['x' + str(lev) + '_' + str(i+1)+'_'+str(j)], 'beta' + str(lev), 'i', lvars['lenbeta' + str(lev)])}); 
<% row += 1 %>\\
<% lvars['lenbeta' + str(lev)] += len(context['x' + str(lev) + '_' + str(i + 1) + '_'+ str(j)]) %>\\
% endfor
% else:
<% lvars['lenbeta' + str(lev)] += len(context['x' + str(lev) + '_' + str(i + 1)]) %>\\
% endif
% endfor
% endif

% for i in range(0, n1):
            double e${row} = mat_ul${lev}(i, ${i});
<% row += 1 %>\\
% endfor

% for i in range(0, lvars['nr' + str(lev)]):
% for j in range(i, lvars['nr' + str(lev)]):
            sb${lev}(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }

% if lvars['allnormal' + str(lev)]:
        int vwu${lev} = mat_ul${lev}.rows() - ${lvars['nr' + str(lev)] + 1};
        matrix_sym_invinplace(sb${lev});
        mat_d_u${lev} = dwishart(vwu${lev}, sb${lev}, fail);
        mat_omega_u${lev} = matrix_sym_inverse(mat_d_u${lev}, fail);
% else:
        if (runstate == 0)
            mat_d_u${lev} = matrix_sym_inverse(mat_omega_u${lev}, fail);

        // Now loop over elements of va (omega_e)
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, lvars['nr' + str(lev)]):
<% var2 = var1 %>\\
% if var1 >= lvars['nvar' + str(lev)]:
<% var2 = var1 + 1%>
% endif
<% withinvar2 = withinvar1 %>\\
% for j in range(i, lvars['nr' + str(lev)]):
        {
% if var1 == var2 and (context['bin' + str(lev) + '_' + str(var1+1)] != 'Normal'):
            omega_u${lev}_accept[${count}] = 44;
% else:
            double llcurr = mat_ul${lev}.rows()*log(matrix_determinant(mat_d_u${lev}))/2;
            for(int i1 = 0; i1 < ${lvars['nr' + str(lev)]}; i1++)
                for(int j1 = 0; j1 < ${lvars['nr' + str(lev)]}; j1++)
                    llcurr -= sb${lev}(i1, j1) * mat_d_u${lev}(i1, j1)/2;
            double old = mat_omega_u${lev}(${i}, ${j});
            mat_omega_u${lev}(${i}, ${j}) = dnorm(mat_omega_u${lev}(${i}, ${j}), 1.0/sqrt(omega_u${lev}_sd[${count}]));
            // mat_d_u${lev} = matrix_sym_inverse(mat_omega_u${lev}, fail);
            SymMatrix temp_d_u${lev} = matrix_cholesky(mat_omega_u${lev}, fail);
            if (fail == true) {
                mat_omega_u${lev}(${i}, ${j}) = old;
                // mat_d_u${lev} = matrix_sym_inverse(mat_omega_u${lev}, fail);
            } else {
                temp_d_u${lev} = matrix_cholinverse(temp_d_u${lev});
                double llnew = mat_ul${lev}.rows()*log(matrix_determinant(temp_d_u${lev}))/2;
                for(int i1 = 0; i1 < ${lvars['nr' + str(lev)]}; i1++)
                    for(int j1 = 0; j1 < ${lvars['nr' + str(lev)]}; j1++)
                        llnew -= sb${lev}(i1, j1) * temp_d_u${lev}(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                    if(random01() > exp(llnew - llcurr))
                        change = 0;
                }
                if(change == 0) {
                    mat_omega_u${lev}(${i}, ${j}) = old;
                    //mat_d_u${lev} = matrix_sym_inverse(mat_omega_u${lev}, fail);
                } else {
                    omega_u${lev}_accept[${count}]++;
                    mat_d_u${lev} = temp_d_u${lev};
                }
            }
% endif
<% count += 1 %>\\
        }
% if var2 < len(context['y' + str(lev)]):

% if (context['bin' + str(lev) + '_' + str(var2+1)] != 'Unordered'):
<% var2 += 1 %>\\
% elif withinvar2 != (context['ncats'+context['y' + str(lev)][var2]]-2):
<% withinvar2 += 1 %>\\
% else:
<% var2 += 1 %>\\
<% withinvar2 = 0 %>\\
% endif

% endif

% endfor

% if var1 < lvars['nvar' + str(lev)]:

% if (context['bin' + str(lev) + '_' + str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+context['y' + str(lev)][var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif

% endif

% endfor

% endif

% endfor



                // Step for beta
% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
% if condmarg:
                // Calculate submatrix of omega_u and invert
		SymMatrix mat_omega_beta${lev}(omega_u${lev}, ${lvars['n' + str(lev)]});
                SymMatrix mat_d_beta${lev} = matrix_sym_inverse(mat_omega_beta${lev}, fail);
% endif

                SymMatrix db${lev}(${lvars['lenbeta' + str(lev)]});
                for (int i = 0; i < length(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][0]}); i++) {
                    for (int j1 = 0; j1 < ${lvars['lenbeta' + str(lev)]}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
% if condmarg:
                            db${lev}(j1, j2) += mat_x${lev}(i, j1) * mat_d_beta${lev}(betaind${lev}[j1], betaind${lev}[j2]) * mat_x${lev}(i, j2);
% else:
                            db${lev}(j1, j2) += mat_x${lev}(i, j1) * mat_d_u${lev}(betaind${lev}[j1], betaind${lev}[j2]) * mat_x${lev}(i, j2);
% endif
                        }
                    }
                }

                RectMatrix xtdy${lev}(${lvars['lenbeta' + str(lev)]}, 1);
                for (int i = 0; i < length(${context['L' + str(lev) + 'Data']}_${context['y' + str(lev)][0]}); i++) {
                    for (int j1 = 0; j1 < ${lvars['lenbeta' + str(lev)]}; j1++) {
                        for (int j2 = 0; j2 < ${lvars['n' + str(lev)]}; j2++) {
% if condmarg:
                            xtdy${lev}(j1, 0) += mat_x${lev}(i, j1) * mat_d_beta${lev}(betaind${lev}[j1], j2) * mat_ymiss${lev}(i, j2);
% else:
                            xtdy${lev}(j1, 0) += mat_x${lev}(i, j1) * mat_d_u${lev}(betaind${lev}[j1], j2) * mat_ymiss${lev}(i, j2);
% endif
                        }
% if not condmarg:
                        for (int j3 = 0; j3 < ${n1}; j3++) {
                            xtdy${lev}(j1, 0) += mat_d_u${lev}(betaind${lev}[j1], j3 + ${lvars['n' + str(lev)]}) * mat_ul${lev}(i, j3);
                        }
% endif
                    }
                }

                fail = false;
                // int err${lev} = matrix_sym_invinplace(db${lev});
                // RectMatrix betahat${lev} = db${lev} * xtdy${lev};
                // mat_beta${lev} = dmultnormal(betahat${lev}, db${lev}, fail);

                RectMatrix betahat${lev} = matrix_solve(db${lev}, xtdy${lev});
                mat_beta${lev} = dmultnormalinv(betahat${lev}, db${lev}, fail);
% endif
% endfor

% for lev in range(2, endlev):
% if context['anyL'+str(lev)+'resp']:
% for i in range(0, lvars['n' + str(lev)]):

<% 
mat_x_i = 'mat_x' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i] + 1)
if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i] + 1)] == 'Unordered'):
    mat_x_i += '_' + str(i - lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]])
%>


                for (int j = 0; j < mat_ymiss${lev}.rows(); j++) {
                    //if (mat_y(j, ${lvars['pyyind' + str(lev)][i]}) <= -9.999e29) {
                    {
                        double tmp = 0.0;
                        for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                            tmp += mat_beta${lev}(betamap${lev}[${i}] + k, 0) * ${mat_x_i}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss${lev}.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_u${lev}(${i}, k) * mat_ymiss${lev}(j, k) * (1.0 / mat_d_u${lev}(${i}, ${i}));
                            }
                        }
% for m in range(0, lvars['n' + str(lev)]):
% if i != m:
<%
mat_x_m = 'mat_x' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][m] + 1)
if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][m] + 1)] == 'Unordered'):
    mat_x_m += '_' + str(m - lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][m]])
%>
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp += mat_d_u${lev}(${i}, ${m}) * (1 / mat_d_u${lev}(${i}, ${i})) * mat_beta${lev}(betamap${lev}[${m}] + k, 0) * ${mat_x_m}(j, k);
                        }
% endif
% endfor

% for m in range(0, n1):
                        tmp -= mat_d_u${lev}(${i}, ${lvars['n' + str(lev)]+m}) * (1 / mat_d_u${lev}(${i}, ${i})) * mat_ul${lev}(j, ${m});
% endfor


% if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i]+1)] == 'Normal'):
                    if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) <= -9.999e29) {
                        mat_ymiss${lev}(j, ${i}) = dnorm(tmp, mat_d_u${lev}(${i}, ${i}));
                    }
% endif
% if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i]+1)] == 'Binary'):
                    if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) <= -9.999e29) 
                        mat_ymiss${lev}(j, ${i}) = dnorm(tmp, mat_d_u${lev}(${i}, ${i}));
                    else if(mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) <= 0) 
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_u${lev}(${i}, ${i})), 2, 0, 0);
                    else          
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_u${lev}(${i}, ${i})), 1, 0, 0);
% endif
% if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i]+1)] == 'Ordered'):
                    if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) <= -9.999e29) 
                        mat_ymiss${lev}(j, ${i}) = dnorm(tmp, mat_d_u${lev}(${i}, ${i}));
                    else if(mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) == 0) 
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 2, 0, 0);
                    else if(mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) == ${context['ncats'+context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]] - 1})
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 1, alpha${context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]}[ncats${context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]}-2], 0);
                    else          
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 3, alpha${context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]}[int(mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}))-1], alpha${context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]}[int(mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}))]);

% endif

% if (context['bin' + str(lev) + '_' + str(lvars['pyyind' + str(lev)][i]+1)] == 'Unordered'):
                    if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) <= -9.999e29) 
                        mat_ymiss${lev}(j, ${i}) = dnorm(tmp, mat_d_u${lev}(${i}, ${i}));
                    else if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) == ${context['ncats'+context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]] - 1}) 
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 2, 0, 0);
                    else if (mat_y${lev}(j, ${lvars['pyyind' + str(lev)][i]}) == ${i - lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]}) {
                        double limit = 0;
                        for (int k = 0; k < ${context['ncats'+context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]] - 1}; k++) {
                            if ((k != ${i - lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]}) && (mat_ymiss${lev}(j, ${lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]} + k) > limit))
                                limit = mat_ymiss${lev}(j, ${lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]} + k); 
                        }
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 1, limit, 0);
                    } else {  
                        double limit = mat_ymiss${lev}(j, ${lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]});
                        for (int k = 0; k < ${context['ncats'+context['y' + str(lev)][lvars['pyyind' + str(lev)][i]]] - 1}; k++) {
                            if (mat_ymiss${lev}(j, ${lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]} + k) > limit)
                                limit = mat_ymiss${lev}(j, ${lvars['pyymap' + str(lev)][lvars['pyyind' + str(lev)][i]]} + k); 
                        }
                        mat_ymiss${lev}(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_u${lev}(${i}, ${i})), 2, 0, limit);
                    }

% endif

                    }
                }
% endfor
% endif
% endfor
        }
    }
'''

    preparedata = '''
import numpy
l1data = data['datafile']
for i in range(0, len(y1)):
    if (context['bin1_'+str(i+1)] == 'Normal'):
        context['miss'+y1[i]].ncols = -1*len(l1data.variables[y1[i]]['data'])
    if (context['bin1_'+str(i+1)] == 'Binary'):
        context['misb'+y1[i]].ncols = -1*len(l1data.variables[y1[i]]['data'])
    if (context['bin1_'+str(i+1)] == 'Ordered'):
        context['miso'+y1[i]].ncols = -1*len(l1data.variables[y1[i]]['data'])
    if (context['bin1_'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+y1[i]]-1):
            context['misu'+str(context['ncats'+y1[i]])+'_'+y1[i]+'_'+str(j)].ncols = -1*len(l1data.variables[y1[i]]['data'])
        context['origv1_'+str(i+1)].name.append(y1[i]) 
        for var in context['origv1_'+str(i+1)].name:
            uniqvals = list(set(l1data.variables[var]['data'].compressed()))
            uniqvals.sort()
            uniqvals.remove(uniqvals[len(uniqvals)-1])
            for j in range(0, len(uniqvals)):
                l1data.addvariable(var + '_' + str(int(j)), data = (numpy.array(l1data.variables[var]['data'])[:] == uniqvals[j]).astype(float))
                context['y1_'+str(i+1)].name.append(var + '_' + str(int(j)))

endlev = Nhigherlevels + 2
for lev in range(2, endlev):
    if context['anyL'+str(lev)+'resp']:
        lhdata = data[str(context['L'+str(lev)+'Data'])]

        for i in range(0, len(context['y'+str(lev)])):
            if (context['bin'+str(lev)+'_'+str(i+1)] == 'Normal'):
                context['miss'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]].ncols = -1*len(numpy.unique(l1data.variables[str(context['L'+str(lev)+'ID'])]['data']))
            if (context['bin'+str(lev)+'_'+str(i+1)] == 'Binary'):
                context['misb'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]].ncols = -1*len(numpy.unique(l1data.variables[str(context['L'+str(lev)+'ID'])]['data']))
            if (context['bin'+str(lev)+'_'+str(i+1)] == 'Ordered'):
                context['miso'+str(context['L'+str(lev)+'Data'])+'_'+context['y'+str(lev)][i]].ncols = -1*len(numpy.unique(l1data.variables[str(context['L'+str(lev)+'ID'])]['data']))
            if (context['bin'+str(lev)+'_'+str(i+1)] == 'Unordered'):
                for j in range(0, context['ncats'+context['y'+str(lev)][i]]-1):
                    context['misu'+str(context['L'+str(lev)+'Data'])+'_'+str(context['ncats'+context['y'+str(lev)][i]])+'_'+context['y'+str(lev)][i]+'_'+str(j)].ncols = -1*len(numpy.unique(l1data.variables[str(context['L'+str(lev)+'ID'])]['data']))

                context['origv'+str(lev)+'_'+str(i+1)].name.append(context['y'+str(lev)][i]) 
                for var in context['origv'+str(lev)+'_'+str(i+1)].name:
                    uniqvals = list(set(lhdata.variables[var]['data'].compressed()))
                    uniqvals.sort()
                    uniqvals.remove(uniqvals[len(uniqvals)-1])
                    for j in range(0, len(uniqvals)):
                        lhdata.addvariable(var + '_' + str(int(j)), data = (numpy.array(lhdata.variables[var]['data'])[:] == uniqvals[j]).astype(float))
                        context['y'+str(lev)+'_'+str(i+1)].name.append(var + '_' + str(int(j)))
'''
